#!/usr/bin/env python3
"""
JGDV, my kitchen sink library.


"""
__version__ = "1.2.0"

from ._abstract import protocols as protos
from ._abstract.types import *  # noqa: F403
from ._abstract import prelude
from . import errors
from .errors import JGDVError
from .decorators import Mixin, Proto

# Subpackage Accessors
from ._abstract import types as Types # noqa: N812
import jgdv.decorators as Decos  # noqa: N812

def identity_fn[T](x:T) -> T:
    """ Just returns what it gets """
    return x



#!/usr/bin/env python3
"""


"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

MODULE_NAME              : Final[str]  = "__module__"
SLOTS_NAME               : Final[str]  = "__slots__"
ANNOTS_NAME              : Final[str]  = "__annotations__"

FreshKWD                 : Final[str]  = "fresh_registry"
Default_K                : Final[str]  = "__default__"
AnnotateKWD              : Final[str]  = "_annotate_to"
AnnotationTarget         : Final[str]  = "__jgdv_typevar__"
AnnotateRx               : Final[Rx]   = re.compile(r"(?P<name>\w+)(?:<(?P<extras>.*?)>)?(?:\[(?P<params>.*?)\])?")

MultiParamFail           : Final[str]  = "Multi Param Annotation not supported yet"
BadParamFail             : Final[str]  = "Bad param value for making an annotated subclass"
NoPydanticFail           : Final[str]  = "Extending pydantic classes with a new mro is not implemented"

NoNameMatch              : Final[str]  = "Couldn't even match the cls name"
NoSubName                : Final[str]  = "No decorated name available"
BadDecorationNameTarget  : Final[str]  = "Unexpected name decoration target"
UnexpectedMRO            : Final[str]  = "Unexpected mro type"
UnexpectedNameSpace      : Final[str]  = "Unexpected namespace type"

ORIG_BASES_K             : Final[str]  = "__orig_bases__"



"""
jgdv.cli provides a statemachine based argument parser.

ParseMachineBase defines the state flow,
ParseMachine implements the __call__ to start the parsing, 
CLIParser implements the callbacks for the different states.

ParamSpec's are descriptions of a single argument type, 
combined with the parsing logic for that type.
"""
from ._interface import ParamStruct_p, ArgParser_p, ParamSource_p, CLIParamProvider_p
from .errors import ParseError
from .arg_parser import ParseMachine
from .param_spec import ParamSpec
from .builder_mixin import ParamSpecMaker_m



#!/usr/bin/env python3
"""

"""
# ruff: noqa: PLW0211, ARG004

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from .strang import Strang
from . import errors

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol

if typing.TYPE_CHECKING:
    from uuid import UUID
    from . import _interface as API # noqa: N812

    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class StrangDebug:

    @staticmethod
    def debug(cls:type[API.Strang_p], text:str) -> str:
        """ Return a str of details about a str -> strang """
        hooks         = [x for x in dir(cls) if x.endswith("_h")]
        abs_sections  = StrangDebug._describe_section_specs(cls._sections)
        ang           = cls(text)
        sections      = StrangDebug._describe_text(cls, ang)
        raw_data      = StrangDebug._describe_raw_data(cls, ang)
        result        = [
            "",
            f"** Strang Debug of: {text} **",
            f"Type: {cls.__module__} : {cls.__qualname__}",
            f"Annotation: {cls.cls_annotation()}",
            f"Hooks: {hooks}",
            *abs_sections,
            *StrangDebug._describe_pre_process_results(cls, text),
            *StrangDebug._describe_process_results(cls, ang),
            *StrangDebug._describe_post_process_results(cls, ang),
            *sections,
            *raw_data,
            "----",
            "Result:",
            f"type(ang)  = {type(ang)}",
            f"str(ang)   = {ang:s}",
            f"ang[:]     = {ang[:]}",
            f"ang[:,:]   = {ang[:,:]}",
            "Formatting: ",
            f"ang:u     = {ang:u}",
            f"ang:a-    = {ang:a-}",
            f"ang:a     = {ang:a}",
            f"ang:a+    = {ang:a+}",
            f"{text} -> {ang}",
        ]
        return "\n".join(result)

    @staticmethod
    def _describe_pre_process_results(cls:type[API.Strang_p], text:str) -> list[str]:
        preproc  = cls._processor.pre_process(cls, text)
        result   = [
            "---- Pre-Process:",
            f"{text} -> {preproc}",
            "",
        ]
        return result

    @staticmethod
    def _describe_process_results(cls:type[API.Strang_p], actual:API.Strang_p) -> list[str]:
        section_slices  = [(x.start, x.stop) for x in actual.data.sections]
        word_slices     = [(x.start, x.stop) for x in actual.data.words]
        result          = [
            "---- Process: ",
            f"Sections:   {section_slices}",
            f"SecWords:   {actual.data.sec_words}",
            f"FlatIdx:    {actual.data.flat_idx}",
            f"Words:      {word_slices}",
            "",
        ]
        return result

    @staticmethod
    def _describe_post_process_results(cls:type[API.Strang_p], actual:API.Strang_p) -> list[str]:
        result = [
            "---- Post-Process:",

            "",
        ]
        return result

    @staticmethod
    def _describe_section_specs(secs:API.Sections_d) -> list[str]:
        result = [
            "---- Section Specs:",
            *(f"- {x.idx} : {x.name}. Case: ({x.case} {x.end}). " for x in secs),
            "",
        ]
        return result

    @staticmethod
    def _describe_text(cls:type[API.Strang_p], actual:API.Strang_p) -> list[str]:
        result = [
            "---- Section Text:",
            "Format: ..-(index[sec], index[word]) : FlatIndex : word : (meta, type[meta])",
        ]

        for i,sec in enumerate(actual.data.sec_words):
            result.append(f"- {actual.section(i).name}:")
            for j,idx in enumerate(sec):
                sl    = actual.data.words[idx]
                word  = actual[sl]
                match actual.data.meta[idx]:
                    case None:
                        meta = ""
                    case x:
                        meta = f"({x}, {type(x)})"

                result.append(f"..- ({i},{j}) : {idx} : {word} : {meta}")
        else:
            args_text       = actual[actual.data.args_start:]
            arg_len  : int  = len(actual.data.args) if actual.data.args else 0
            result += [
                "",
                f"- args ({arg_len}). Position: {actual.data.args_start}:",
                f"..- {args_text}",
                "",
            ]
            return result


    @staticmethod
    def _describe_raw_data(cls:type[API.Strang_p], actual:API.Strang_p) -> list[str]:
        result = [
            "---- Raw Data:",
            f"Meta:    {actual.data.meta}",
            f"UUID:    {actual.data.uuid}",
            "",
            ]
        return result

    @staticmethod
    def diff_uuids(this:UUID, that:UUID) -> str:
        this_ing : str = str(this)
        that_ing : str = str(that)
        result = [y if x==y else "_" for x,y in zip(this_ing, that_ing, strict=True)]
        return "".join(result)



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast, Self, Final)
import warnings

import pytest

logging = logmod.root

from jgdv.structs.strang import CodeReference
from ... import DKey
from ..._interface import Key_p
from ..import_key import ImportDKey

IMP_KEY_BASES               : Final[list[str]]           = ["bob", "bill", "blah", "other", "23boo", "aweg2531", "awe_weg", "aweg-weji-joi"]
EXP_KEY_BASES               : Final[list[str]]           = [f"{{{x}}}" for x in IMP_KEY_BASES]
EXP_P_KEY_BASES             : Final[list[str]]           = ["{bob:wd}", "{bill:w}", "{blah:wi}", "{other:i}"]
PATH_KEYS                   : Final[list[str]]           = ["{bob}/{bill}", "{blah}/{bloo}", "{blah}/{bloo}"]
MUTI_KEYS                   : Final[list[str]]           = ["{bob}_{bill}", "{blah} <> {bloo}", "! {blah}! {bloo}!"]
IMP_IND_KEYS                : Final[list[str]]           = ["bob_", "bill_", "blah_", "other_"]
EXP_IND_KEYS                : Final[list[str]]           = [f"{{{x}}}" for x in IMP_IND_KEYS]

VALID_KEYS                                           = IMP_KEY_BASES + EXP_KEY_BASES + EXP_P_KEY_BASES + IMP_IND_KEYS + EXP_IND_KEYS
VALID_MULTI_KEYS                                     = PATH_KEYS + MUTI_KEYS

class TestImportKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_str_mark(self):
        assert(DKey.MarkOf(ImportDKey) is CodeReference)

    def test_basic(self):
        match DKey[CodeReference]("fn", implicit=True):
            case ImportDKey():
                assert(True)
            case x:
                 assert(False), x


    def test_expand(self):
        key = DKey[CodeReference]("fn", implicit=True)
        match key({"fn":"jgdv:identity_fn"}):
            case CodeReference() as x:
                iden = x()
                assert(callable(iden))
                assert(iden(2) == 2)
            case x:
                 assert(False), x


    def test_expand_annotated(self):
        key = DKey[CodeReference]("fn", implicit=True)
        match key({"fn":"fn::jgdv:identity_fn"}):
            case CodeReference() as x:
                assert(callable(x()))
                assert(x()(2) == 2)
                assert(True)
            case x:
                 assert(False), x

    def test_expand_redirect(self):
        key = DKey[CodeReference]("fn", implicit=True)
        match key({"fn_": "acceptor", "acceptor": "fn::jgdv:identity_fn"}):
            case CodeReference() as x:
                assert(callable(x()))
                assert(x()(2) == 2)
                assert(True)
            case x:
                 assert(False), x



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from collections import defaultdict
from uuid import UUID, uuid1

# ##-- end stdlib imports

from . import _interface as API # noqa: N812
from .tag_file import TagFile

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class SubstitutionFile(TagFile):
    """ SubstitutionFiles add a replacement tag for some tags

    Substitution file format is single lines of:
    ^{tag} {sep} {count} [{sep} {replacement}]*$

    """

    sep           : str                  = API.SEP
    ext           : str                  = API.SUB_EXT
    substitutions : dict[str, set[str]]  = defaultdict(set)  # noqa: RUF012

    def __str__(self) -> str:
        """
        Export the substitutions, 1 entry per line, as:
        `key` : `counts` : `substitution`
        """
        all_lines = []
        for key in sorted(self.counts.keys()):
            if not bool(self.substitutions[key]):
                continue
            line = [key, str(self.counts[key])]
            line += sorted(self.substitutions[key])
            all_lines.append(self.sep.join(line))

        return "\n".join(all_lines)

    def __getitem__(self, key:str) -> set[str]:
        """ Gets the substitutions for a key """
        return self.sub(key)

    def canonical(self) -> TagFile:
        """ create a tagfile of just canonical tags"""
        # All substitutes are canonical
        canon = {x:1 for x in iter(self) if not self.has_sub(x)}
        return TagFile(counts=canon)

    def known(self) -> TagFile:
        """ Get a TagFile of all known tags. both canonical and not """
        canon = self.canonical()
        canon += self
        return canon

    def sub(self, value:str) -> set[str]:
        """ apply a substitution if it exists """
        normed = self.norm_tag(value)
        if bool(self.substitutions.get(normed, None)):
            return self.substitutions[normed]

        return {normed}

    def sub_many(self, *values:str) -> set[str]:
        result = set()
        for val in values:
            result.update(self.sub(val))
        else:
            return result

    def has_sub(self, value:str) -> bool:
        normed = self.norm_tag(value)
        if normed != value:
            return True
        return bool(self.substitutions.get(normed, None))

    def update(self, *values:str|tuple|dict|SubstitutionFile|TagFile|set) -> Self:  # noqa: PLR0912
        """
        Overrides TagFile.update to handle tuples of (tag, count, replacements*)
        and (tag, replacements*)
        """
        for val in values:
            match val:
                case None | "": # empty line
                    continue
                case str() if val.startswith(self.comment): # comment
                    continue
                case str() if API.ALT_SEP in val: # key :: subs
                    x, *ys = val.split(API.ALT_SEP)
                    self._add_sub(x, list(ys))
                case str() if self.sep in val: # key : count : subs
                    self.update(tuple(x.strip() for x in val.split(self.sep)))
                case str(): # key
                    self._inc(val)
                case list() | set(): # keys
                    for key in val:
                        self._inc(key)
                case dict(): # { key : count }
                    for key, v in val.items():
                        self._inc(key, amnt=v)
                case (str() as key, int() | str() as counts): # (key, count)
                    self._inc(key, amnt=int(counts))
                case (str() as key, list() as subs): # (key, subs)
                    self._add_sub(key, subs)
                case (str() as key, int() | str() as counts, *subs): # key, count, subs
                    self._add_sub(key, list(subs), count=counts)
                case SubstitutionFile() as v:
                    self.update(v.counts)
                    for tag, xs in v.substitutions.items():
                        self._add_sub(tag, xs)
                case TagFile():
                    self.update(*val.counts.items())

        return self

    def _add_sub(self, key:str, subs:Iterable[str], *, count:str|int=1) -> None:
        match subs:
            case [str() as x] if self.sep in x:
                subs = x.split(self.sep)
            case _:
                pass
        try:
            count = int(count)
        except ValueError:
            count = 1

        match self._inc(key, amnt=count):
            case None:
                pass
            case str() as norm_key:
                norm_subs = [normed for x in subs if (normed:=self.norm_tag(x)) is not None]
                self.update(dict.fromkeys(norm_subs, 1)) # Add to normal counts too
                self.substitutions[norm_key].update(norm_subs)



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import uuid
import warnings
from collections.abc import (Callable, Iterable, Iterator, Mapping,
                             MutableMapping, Sequence)
from random import randint
from re import Match
from typing import Annotated, Any, ClassVar, Generic, TypeAlias, TypeVar, cast
from uuid import UUID
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

from .. import _interface as API  # noqa: N812
from ..errors import StrangError
from ..processor import StrangBasicProcessor
from ..strang import Strang

##--|
logging  = logmod.root
UUID_STR = str(uuid.uuid1())

##--|

class TestStrang_PreProcess:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_clean_separators(self):
        ing =  "a.b.c::d.e....f"
        obj = StrangBasicProcessor()
        match obj.pre_process(Strang, ing):
            case "a.b.c::d.e..f", {}, {}, None:
                assert(True)
            case x:
                assert(False), x

    def test_trim_rhs(self):
        ing = "a.b.c::d.e...."
        obj = StrangBasicProcessor()
        match obj.pre_process(Strang, ing):
            case "a.b.c::d.e", {}, {}, None:
                assert(True)
            case x:
                 assert(False), x

    def test_trim_lhs(self):
        ing =  "    a.b.c::d.e"
        obj = StrangBasicProcessor()
        match obj.pre_process(Strang, ing):
            case "a.b.c::d.e", {}, {}, None:
                assert(True)
            case x:
                 assert(False), x

    def test_verify_structure_fail_missing(self):
        obj = StrangBasicProcessor()
        with pytest.raises(ValueError):
            obj.pre_process(Strang, "a.b.c")


    def test_compress_uuid(self):
        obj         = StrangBasicProcessor()
        ing         = f"head::a.b.c..<uuid:{UUID_STR}>"
        simple_ing  = "head::a.b.c..<uuid>"
        match obj.pre_process(Strang, ing):
            case str() as x, {}, {"types": [("uuid", str() as uid)]}, None:
                assert(x == simple_ing)
                assert(str(uid) == UUID_STR)
                assert(True)
            case x:
                assert(False), x

    @pytest.mark.xfail
    def test_verify_structure_fail_surplus(self):
        obj = StrangBasicProcessor()
        with pytest.raises(ValueError):
            obj.pre_process(Strang, "a.b.c::d.e.f::g.h.i")


    def test_process_args(self):
        ing =  "a.b.c::d.e.f[<uuid>]"
        obj = StrangBasicProcessor()
        match obj.pre_process(Strang, ing):
            case str() as x, {}, dict() as data, None:
                assert(data.get('args_start', False))
            case x:
                assert(False), x


    def test_process_multi_args(self):
        ing =  "a.b.c::d.e.f[<uuid>,blah,bloo]"
        obj = StrangBasicProcessor()
        match obj.pre_process(Strang, ing):
            case str() as x, {}, dict() as data, None:
                assert(data.get('args_start', False))
            case x:
                assert(False), x


    def test_process_no_args(self):
        ing =  "a.b.c::d.e.f"
        obj = StrangBasicProcessor()
        match obj.pre_process(Strang, ing):
            case str() as x, {}, dict() as data, None:
                assert("args_start" not in data)
            case x:
                assert(False), x

class TestStrang_Process:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_process_section_bounds(self):
        ing                 = "a.b.c::d.e.f"
        obj                 = StrangBasicProcessor()
        base                = Strang(ing)
        base.data.sections  = ()
        assert(not bool(base.data.sections))
        obj.process(base, data={})
        assert(bool(base.data.sections))
        match base.data.sections:
            case [slice() as head, slice() as body]:
                assert(ing[head] == "a.b.c")
                assert(ing[body] == "d.e.f")
                assert(True)
            case x:
                 assert(False), x

    def test_process_words(self):
        ing              = "a.b.c.blah::d.e.f"
        obj              = StrangBasicProcessor()
        base             = Strang(ing)
        words            = ["a","b","c","blah","d","e","f"]
        base.data.words  = ()
        assert(not bool(base.data.words))
        obj.process(base, data={})
        assert(bool(base.data.words))
        match base.data.words:
            case [*xs]:
                assert(len(xs) == 7)
                for x,sl in zip(words, xs, strict=True):
                    assert(x == ing[sl])
            case x:
                 assert(False), x


    def test_process_word_idxs(self):
        ing              = "a.b.c.blah::d.e.f"
        obj              = StrangBasicProcessor()
        base             = Strang(ing)
        target_words            = ["a","b","c","blah","d","e","f"]
        base.data.sec_words = ()
        assert(not bool(base.data.sec_words))
        obj.process(base, data={})
        assert(bool(base.data.sec_words))
        actual_words = [base.data.words[y] for x in base.data.sec_words for y in x]
        for x,sl in zip(target_words, actual_words, strict=True):
            assert(x == ing[sl])

    def test_process_section_flat(self):
        ing              = "a.b.c.blah::d.e.f"
        words            = ["a","b","c","blah","d","e","f"]
        obj              = StrangBasicProcessor()
        base             = Strang(ing)
        base.data.words  = ()
        assert(not bool(base.data.words))
        obj.process(base, data={})
        match base.data.words:
            case [*xs]:
                assert(len(xs) == 7)
                for x,sl in zip(words, xs, strict=True):
                    assert(x == ing[sl])
            case x:
                assert(False), x


    def test_process_args(self):
        ing             = "a.b.c.blah::d.e.f[blah,bloo,blee]"
        words           = ["a","b","c","blah","d","e","f"]
        obj             = StrangBasicProcessor()
        ang             = Strang(ing)
        ang.data.args = None
        obj.process(ang, data={})
        assert(ang.data.args_start == ing.rindex(API.ARGS_CHARS[0]))
        assert(ang.data.sections[-1].stop <= ing.rindex(API.ARGS_CHARS[0]))
        assert(bool(ang.data.args))
        assert(set(ang.data.args) == set(["blah","bloo","blee"]))


    def test_process_args_set_uuid(self):
        ing                         = "a.b.c.blah::d.e.f[<uuid>]"
        words                       = ["a","b","c","blah","d","e","f"]
        obj                         = StrangBasicProcessor()
        _, _, post_data, _  = obj.pre_process(Strang, ing)
        ang                         = Strang(ing)
        ang.data.args               = None
        obj.process(ang, data=post_data)
        assert(ang.data.args_start  == ing.rindex(API.ARGS_CHARS[0]))
        assert(ang.data.sections[-1].stop <= ing.rindex(API.ARGS_CHARS[0]))
        assert(bool(ang.data.args))
        assert(set(ang.data.args) == set(["<uuid>"]))
        assert(isinstance(ang.data.uuid, UUID))


class TestStrang_PostProcess_UUIDs:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_head_uuids(self):
        obj = StrangBasicProcessor()
        val = Strang("a.b.c.<uuid>::d.e.f")
        val.data.uuid = None
        val.data.meta = ()
        assert(not bool(val.data.meta))
        assert(not val.uuid())
        obj.post_process(val, {"types": [("uuid", None)]})
        assert(bool(val.data.meta))
        assert(not val.uuid())
        match val.data.meta[3]:
            case UUID():
                assert(True)
            case x:
                assert(False), x

    def test_body_uuids(self):
        obj = StrangBasicProcessor()
        val = Strang("a.b.c::d.e.<uuid>")
        assert(not val.uuid())
        val.data.meta = ()
        assert(not bool(val.data.meta))
        obj.post_process(val, {"types" : [("uuid", None)]})
        assert(not val.uuid())
        match val.data.meta[-1]:
            case UUID():
                assert(True)
            case x:
                 assert(False), x

    def test_exact_uuid(self):
        obj = Strang(f"head::tail.<uuid:{UUID_STR}>")
        assert(not obj.uuid())
        match obj.get(1,-1):
            case uuid.UUID():
                assert(True)
            case x:
                 assert(False), type(x)

    def test_rebuild_uuid(self):
        ing = f"head::tail.<uuid:{UUID_STR}>"
        s1 = Strang(ing)
        s2 = Strang(s1[:,:])
        assert(s1.uuid() == s2.uuid())
        assert(isinstance(s1.get(1,-1), uuid.UUID))
        assert(isinstance(s2.get(1,-1), uuid.UUID))
        assert(s1.get(1,-1) == s2.get(1,-1))
        assert(s1[1,-1] == s2[1,-1])

    def test_rebuild_generated_uuid(self):
        s1 = Strang("head::tail.<uuid>")
        s2 = Strang(s1[:,:])
        assert(not s1.uuid())
        assert(not s1.uuid())
        assert(isinstance(s1.get(1,-1), uuid.UUID))
        assert(isinstance(s2.get(1,-1), uuid.UUID))
        assert(s1[:,:] == s2[:,:])

    def test_too_many_uuids(self):
        with pytest.raises(StrangError):
            Strang("head::tail[<uuid>,<uuid>]")


    def test_multiple_uuids(self):
        obj = Strang("head::tail.<uuid>.<uuid>")
        match obj.get(1,1), obj.get(1,2):
            case UUID(), UUID():
                assert(True)
            case x:
                assert(False), x

class TestStrang_PostProcess_Marks:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_head_marks(self):
        head           = API.DefaultHeadMarks_e
        obj            = StrangBasicProcessor()
        val            = Strang("a.b.$basic$::d.e.f")
        val.data.meta  = ()
        assert(not bool(val.data.meta))
        obj.post_process(val, {})
        assert(bool(val.data.meta))
        match val.data.meta[2]:
            case x if x is head.basic:
                assert(val.get(0,-1) is head.basic)
                assert(True)
            case x:
                assert(False), x

    def test_body_marks(self):
        body = API.DefaultBodyMarks_e
        obj = StrangBasicProcessor()
        val = Strang("a.b.c::d.e.$head$")
        val.data.meta = ()
        assert(not bool(val.data.meta))
        obj.post_process(val, {})
        assert(bool(val.data.meta))
        match val.data.meta[-1]:
            case x if x is body.head:
                assert(val.get(1,-1) is body.head)
                assert(True)
            case x:
                assert(False), x

    def test_implicit_mark_start(self):
        body = API.DefaultBodyMarks_e
        obj = StrangBasicProcessor()
        val = Strang(f"head::_.tail.blah")
        val.data.meta = ()
        assert(not bool(val.data.meta))
        obj.post_process(val, {})
        assert(bool(val.data.meta))
        match val.data.meta[1]:
            case x if x is body.hide:
                assert(val.get(1,0) is body.hide)
            case x:
                assert(False), x

    def test_implicit_mark_end(self):
        body           = API.DefaultBodyMarks_e
        obj            = StrangBasicProcessor()
        val            = Strang(f"head::tail.blah._")
        val.data.meta  = ()
        assert(not bool(val.data.meta))
        obj.post_process(val, {})
        assert(bool(val.data.meta))
        match val.data.meta[-1]:
            case x if x is body.hide:
                assert(val.get(1,-1) is body.hide)
            case x:
                assert(False), x

    def test_implicit_skip_mark(self):
        body           = API.DefaultBodyMarks_e
        obj            = StrangBasicProcessor()
        val            = Strang(f"head::tail..blah")
        val.data.meta  = ()
        assert(not bool(val.data.meta))
        obj.post_process(val, {})
        assert(bool(val.data.meta))
        match val.data.meta[2]:
            case x if x is body.empty:
                assert(val.get(1,1) is body.empty)
            case x:
                assert(False), x

    def test_implicit_mark_fail(self):
        """ An implicit mark, not at the start or end, wont be converted """
        body = API.DefaultBodyMarks_e
        obj = StrangBasicProcessor()
        val = Strang(f"head::a._.tail.blah")
        val.data.meta = ()
        assert(not bool(val.data.meta))
        obj.post_process(val, {})
        assert(bool(val.data.meta))
        match val.data.meta[1]:
            case x if x is not body.hide:
                assert(val.get(1,0) is not body.hide)
            case x:
                assert(False), x

    def test_extension_mark(self):
        obj = Strang(f"head::+.tail.blah")
        assert(obj.get(1,0) == Strang.section(1).marks.extend)

    # @pytest.mark.parametrize(["val"], [(x,) for x in iter(Strang.bmark_e)])
    # def test_build_named_mark(self, val):
    #     obj = Strang(f"head::{val}.blah")
    #     assert(obj._body_meta[0] == val)
    #     assert(obj[0] == val)



#!/usr/bin/env python3
"""



"""
# Import:
from __future__ import annotations
from typing import Final, Self
from jgdv.debugging import TraceBuilder

##--| Error Messages
ErrorMSg_1 : Final[str] = "Test Message"


##--| Errors
class JGDVError(Exception):
    """ A Base Error Class for JGDV """

    def __getitem__(self, val:None|int|slice) -> Self:
        """ Use jgdv.debugging.TraceBuilder to control the traceback
        of this error

        """
        match val:
            case int() as x:
                return self.with_traceback(TraceBuilder()[slice(x)])
            case slice(start=None) as x:
                adjusted = slice(1, x.stop, x.step)
                return self.with_traceback(TraceBuilder()[adjusted])
            case slice(start=int() as start) as x:
                adjusted = slice(1+start, x.stop, x.step)
                return self.with_traceback(TraceBuilder()[adjusted])
            case _:
                return self.with_traceback(TraceBuilder()[1:])



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class Hashable_p(Protocol):

    def __hash__(self) -> int: ...

class Iterable_p[V](Protocol):

    def __iter__(self) -> Iterator_p[V]: ...

class Iterator_p[V](Iterable_p, Protocol):

    def __next__(self) -> V: ...

    def __iter__(self) -> Self: ...

class Reversible_p(Iterable_p, Protocol):

    def __reversed__(self) -> Self: ...

class Generator_p(Iterator_p, Protocol):

    def __next__(self) -> Any: ...  # noqa: ANN401

    def send(self, value) -> Any: ...  # noqa: ANN001, ANN401

    def throw(self, typ, val=None, tb=None) -> Any: ...  # noqa: ANN001, ANN401

    def close(self) -> None: ...

class Sized_p(Protocol):

    def __len__(self) -> int: ...

class Container_p[V](Protocol):

    def __contains__(self, x:V) -> bool: ...

class Collection_p(Sized_p, Iterable_p, Container_p, Protocol):
    pass

class Buffer_p(Protocol):

    def __buffer__(self, flags: int) -> memoryview: ...

class Callable_p[*A, K, R](Protocol):

    def __call__(self, *args:*A, **kwds:K) -> R: ...

### SETS ###

class Set_p[V](Collection_p, Protocol):
    """A set is a finite, iterable container.

    This class provides concrete generic implementations of all
    methods except for __contains__, __iter__ and __len__.

    To override the comparisons (presumably for speed, as the
    semantics are fixed), redefine __le__ and __ge__,
    then the other operations will automatically follow suit.
    """

    def __le__(self, other:Self) -> bool: ...

    def __lt__(self, other:Self) -> bool: ...

    def __gt__(self, other:Self) -> bool: ...

    def __ge__(self, other:Self) -> bool: ...

    def __eq__(self, other:Self) -> bool: ... # type: ignore[override]

    @classmethod
    def _from_iterable(cls:type[Set_p], it:Iterable_p[V]) -> Set_p[V]: ...

    def __and__(self, other:Self) -> Self: ...

    def __rand__(self, other:Self) -> Self: ...

    def isdisjoint(self, other:Self) -> bool: ...

    def __or__(self, other:Self) -> Self: ...

    def __ror__(self, other:Self) -> Self: ...

    def __sub__(self, other:Self) -> Self: ...

    def __rsub__(self, other:Self) -> Self: ...

    def __xor__(self, other:Self) -> Self: ...

    def __rxor__(self, other:Self) -> Self: ...

    def _hash(self) -> int: ...

class MutableSet_p[V](Set_p[V], Protocol):
    """A mutable set is a finite, iterable container.

    This class provides concrete generic implementations of all
    methods except for __contains__, __iter__, __len__,
    add(), and discard().

    To override the comparisons (presumably for speed, as the
    semantics are fixed), all you have to do is redefine __le__ and
    then the other operations will automatically follow suit.
    """

    def add(self, value:V) -> Self: ...

    def discard(self, value:V) -> None: ...

    def remove(self, value:V) -> None: ...

    def pop(self) -> V: ...

    def clear(self) -> None: ...

    def __ior__(self, it:Iterable_p) -> Self: ...

    def __iand__(self, it:Iterable_p) -> Self: ...

    def __ixor__(self, it:Iterable_p) -> Self: ...

    def __isub__(self, it:Iterable_p) -> Self: ...

### MAPPINGS ###

class Mapping_p[K,V](Collection_p, Protocol):
    """A Mapping_p is a generic container for associating key/value
    pairs.

    This class provides concrete generic implementations of all
    methods except for __getitem__, __iter__, and __len__.
    """
    __hash__ = None

    def __getitem__(self, key:K) -> V: ...

    def get[D:Maybe](self, key:K, default:Maybe[Any]=None) -> V|D: ...

    def __contains__(self, key:K) -> bool: ...

    def keys(self) -> KeysView_p: ...

    def items(self) -> ItemsView_p: ...

    def values(self) -> ValuesView_p: ...

    def __eq__(self, other:object) -> bool: ...

class MappingView_p(Sized_p, Protocol):

    def __init__(self, mapping:Mapping_p) -> None: ...

    def __len__(self) -> int: ...

    def __repr__(self) -> str: ...

class KeysView_p[K](MappingView_p, Set_p, Protocol):

    @classmethod
    def _from_iterable[V](cls:type[KeysView_p], it:Iterable_p[K]) -> KeysView_p[K]: ...

    def __contains__(self, key:K) -> bool: ...

    def __iter__(self) -> Iterator_p: ...

class ItemsView_p[K,V](MappingView_p, Set_p, Protocol):

    @classmethod
    def _from_iterable(cls:type[ItemsView_p], it:Iterable_p) -> ItemsView_p: ...

    def __contains__(self, item:tuple[K,V]) -> bool: ...

    def __iter__(self) -> Iterator_p[tuple[K,V]]: ...

class ValuesView_p[V](MappingView_p, Collection_p, Protocol):

    def __contains__(self, value:V) -> bool: ...

    def __iter__(self) -> Iterator_p: ...

class MutableMapping_p[K,V](Mapping_p[K,V], Protocol):
    """A MutableMapping is a generic container for associating
    key/value pairs.

    This class provides concrete generic implementations of all
    methods except for __getitem__, __setitem__, __delitem__,
    __iter__, and __len__.
    """

    def __setitem__(self, key:K, value:V) -> None: ...

    def __delitem__(self, key:K) -> None: ...

    def pop[D:V|Maybe[object]](self, key:K, default:Maybe[D]=None) -> D: ...

    def popitem(self) -> V: ...

    def clear(self) -> None: ...

    def update(self, other:Iterable_p=(), /, **kwds:Any) -> None: ...  # noqa: ANN401

    def setdefault[D:V|Maybe](self, key:K, default:Maybe[D]=None) -> V|D: ...

### SEQUENCES ###

class Sequence_p[V](Reversible_p, Collection_p, Protocol):
    """All the operations on a read-only sequence.

    Concrete subclasses must override __new__ or __init__,
    __getitem__, and __len__.
    """

    # Tell ABCMeta.__new__ that this class should have TPFLAGS_SEQUENCE set.

    def __getitem__(self, index:int) -> V: ...

    def __iter__(self) -> Iterator_p: ...

    def __contains__(self, value:V) -> bool: ...

    def __reversed__(self) -> Self: ...

    def index(self, value:V, start:int=0, stop:Maybe[int]=None) -> int: ...

    def count(self, value:V) -> int: ...

class MutableSequence_p[V](Sequence_p, Protocol):
    """All the operations on a read-write sequence.

    Concrete subclasses must provide __new__ or __init__,
    __getitem__, __setitem__, __delitem__, __len__, and insert().
    """

    def __setitem__(self, index:int, value:V) -> None: ...

    def __delitem__(self, index:int) -> None: ...

    def insert(self, index:int, value:V) -> None: ...

    def append(self, value:V) -> None: ...

    def clear(self) -> None: ...

    def reverse(self) -> None: ...

    def extend(self, values:Iterable_p[V]) -> None: ...

    def pop(self, index:int=-1) -> V: ...

    def remove(self, value:V) -> None: ...

    def __iadd__(self, values:Iterable_p[V]) -> Self: ...



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, PLR0133, B011, ANN001, ANN202
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

from .. import param_spec as Specs  # noqa: N812
from ..arg_parser import CLIParser, ParseMachine, ParseResult_d
from ..param_spec import ParamSpec
from ..import param_spec as core

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from dataclasses import InitVar, dataclass, field

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

logging = logmod.root

@dataclass
class _ParamSource:
    name : str
    _param_specs : list[ParamSpec]

    def param_specs(self) -> list[ParamSpec]:
        return self._param_specs

##--|

class TestParseMachine:

    def test_sanity(self):
        assert(True is not False)

    def test_parse_empty(self):
        obj = ParseMachine()
        match obj([], head_specs=[], cmds=[], subcmds=[]):
            case {"head": {}, "cmd": {}, "sub":{}, "extra":{"args":{}, "name": "_extra_"}}:
                assert(True)
            case x:
                assert(False), x

    def test_parse_no_specs(self):
        obj = ParseMachine()
        match obj(["test", "blah"], head_specs=[], cmds=[], subcmds=[]):
            case {"head": {}, "cmd": {}, "sub":{}, "extra":{"args":{}, "name": "_extra_"}}:
                assert(True)
            case x:
                assert(False), x

class TestParser:

    @pytest.fixture(scope="function")
    def a_source(self) -> _ParamSource:

        def builder(name:str) -> ParamSource_p:
            obj = _ParamSource(name=name,
                               _param_specs=[
                                   core.ToggleParam(name="-on", type=bool),
                                   core.KeyParam(name="-val"),
                               ],
                               )
            return obj
        ##--|
        return builder

    def test_sanity(self):
        assert(True is not False)

    def test_setup(self, a_source):
        obj = CLIParser()
        obj._setup(["a","b","c"], [ParamSpec(name="blah")], [a_source("testcmd")], [])
        assert(obj._initial_args == ["a","b","c"])
        assert(obj._remaining_args == ["a","b","c"])
        assert(len(obj._head_specs) == 1)
        assert(obj._head_specs[0].name == "blah")
        assert("testcmd" in obj._cmd_specs)
        assert(not bool(obj._subcmd_specs))

    def test_check_for_help_flag(self):
        obj = CLIParser()
        obj._remaining_args = ["a","b","c", "--help"]
        obj.help_flagged()
        assert(obj._force_help)

    def test_check_for_help_flag_fail(self):
        obj = CLIParser()
        obj._remaining_args = ["a","b","c"]
        obj.help_flagged()
        assert(not obj._force_help)

    def test_parse_empty(self, a_source):
        the_cmd  = a_source("testcmd")
        in_args  = []
        obj      = CLIParser()
        obj._setup(in_args, [ParamSpec(name="blah")], [the_cmd], [])
        obj._parse_head()
        assert(True)

    def test_parse_head(self, a_source):
        the_cmd  = a_source("testcmd")
        in_args  = ["-blah","b","c"]
        obj = CLIParser()
        obj._setup(in_args, [ParamSpec(name="-blah")], [the_cmd], [])
        obj._parse_head()
        assert(obj.head_result.name == "_head_")
        assert(obj.head_result.args['blah'] is True)

    def test_parse_cmd(self, a_source):
        the_cmd  = a_source("testcmd")
        in_args  = ["testcmd", "-val", "aweg", "b","c"]
        obj      = CLIParser()
        obj._setup(in_args, [], [the_cmd], [])
        assert(bool(obj._cmd_specs))
        obj._parse_cmd()
        assert(obj.cmd_result.name == "testcmd")
        match obj.cmd_result.args:
            case {"val": "aweg"}:
                assert(True)
            case x:
                assert(False), x

    def test_parse_cmd_arg_same_as_subcmd(self, a_source):
        the_cmd  = a_source("testcmd")
        in_args = ["testcmd", "-val", "blah", "b","c"]
        obj      = CLIParser()
        obj._setup(in_args, [], [the_cmd], [])
        assert(bool(obj._cmd_specs))
        obj._parse_cmd()
        assert(obj.cmd_result.name == "testcmd")
        match obj.cmd_result.args:
            case {"val": "blah"}:
                assert(True)
            case x:
                 assert(False), x
        assert(not bool(obj.subcmd_results))

    def test_parse_subcmd(self, a_source):
        first_cmd   = a_source("testcmd")
        second_cmd  = a_source("blah")
        in_args     = ["testcmd", "blah"]
        obj         = CLIParser()
        obj._setup(in_args, [], [first_cmd], [(first_cmd.name, second_cmd)])
        assert(obj._subcmd_specs['blah'] == ("testcmd", second_cmd.param_specs()))
        assert(not bool(obj.subcmd_results))
        obj._parse_cmd()
        obj._parse_subcmd()
        assert(obj.cmd_result.name == "testcmd")
        assert(len(obj.subcmd_results) == 1)
        assert(obj.subcmd_results[0].name == "blah")

    def test_parse_multi_subcmd(self, a_source):
        first_cmd             = a_source("testcmd")
        second_cmd            = a_source("blah")
        third_cmd             = a_source("bloo")
        in_args               = ["testcmd", "blah", "-on", "--", "bloo", "-val", "lastval"]
        expected_sub_results  = 2
        obj                   = CLIParser()
        obj._setup(in_args,
                   [],
                   [first_cmd],
                   [(first_cmd.name, second_cmd), (first_cmd.name, third_cmd)])
        assert(obj._subcmd_specs['blah'] == ("testcmd", first_cmd.param_specs()))
        assert(not bool(obj.subcmd_results))
        obj._parse_cmd()
        obj._parse_subcmd()
        assert(obj.cmd_result.name == "testcmd")
        assert(len(obj.subcmd_results) == expected_sub_results)
        assert(obj.subcmd_results[0].name == "blah")
        match obj.subcmd_results[0].args:
            case {"on": True}:
                assert(True)
            case x:
                 assert(False), x

        match obj.subcmd_results[1].args:
            case {"on": False, "val": "lastval"}:
                assert(True)
            case x:
                 assert(False), x

class TestParseResultReport:

    def test_sanity(self):
        assert(True is not False)

    def test_report(self):
        obj = CLIParser()
        obj.head_result = ParseResult_d(name="blah")
        obj.cmd_result = ParseResult_d(name="bloo")
        obj.subcmd_results = []
        match obj.report():
            case {"head": {"name":"blah"}, "cmd": {"name":"bloo"}}:
                assert(True)
            case x:
                 assert(False), x

class TestParseArgs:

    def test_sanity(self):
        assert(True is not False)

    def test_non_positional_params(self):
        params = [
            core.ToggleParam(**{"name":"--aweg", "type":"bool"}),
            core.KeyParam(**{"name":"-val", "type":"str"}),
            core.ToggleParam(**{"name":"-on", "type":"bool"}),
        ]
        obj = CLIParser()
        obj._setup(["--aweg", "-val", "bloo", "-on", "blah"], [],[],[])
        result = ParseResult_d("test results")
        obj._parse_params_unordered(result, params)
        assert(result.args['aweg'] is True)
        assert(result.args['on'] is True)
        assert(result.args['val'] == "bloo")

    def test_positional_params(self):
        params = sorted([
            core.PositionalParam(**{"name":"<4>val",  "type":"str"}),
            core.PositionalParam(**{"name":"<1>blah", "type":"str"}),
            core.PositionalParam(**{"name":"<2>bloo", "type":"str"}),
        ], key=ParamSpec.key_func)
        obj = CLIParser()
        # -on before --aweg
        obj._setup(["first", "second", "third"], [],[],[])
        result = ParseResult_d("test results")
        obj._parse_params_unordered(result, params)
        assert(result.args['blah'] == "first")
        assert(result.args['bloo'] == "second")
        assert(result.args['val'] == "third")



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from copy import deepcopy
from time import sleep
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 1st party imports
from .._interface import TomlTypes
from .._base import GuardBase
from jgdv.structs.chainguard.mixins.reporter_m import DefaultedReporter_m

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
from typing import Any, Never

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class GuardProxy:
    """ A Base Class for Proxies """
    __slots__ = ("_types", "_data", "__index")

    def __init__(self, data:GuardBase, types:Any=None, index:Maybe[list[str]]=None, fallback:TomlTypes|Literal[Never]=Never):
        self._types                         = types or Any
        self._data                          = data
        self.__index : list[str]            = index or ["<root>"]

    def __repr__(self) -> str:
        type_str = self._types_str()
        index_str = ".".join(self._index())
        return f"<GuardProxy: {index_str}:{type_str}>"

    def __len__(self) -> int:
        if hasattr(self._data, "__len__"):
            return len(self._data)

        return 0

    def __bool__(self) -> bool:
        return self._data is not None and self._data is not Never

    def __call__(self, *, wrapper:Maybe[callable[[TomlTypes], Any]]=None, **kwargs) -> Any:
        return None

    def _inject(self, val:tuple[Any]=Never, attr:Maybe[str]=None, clear:bool=False) -> GuardProxy:
        match val:
            case _ if clear:
                val = Never
            case x if x is Never:
                val = self._data
            case _:
                pass

        return GuardProxy(val,
                          types=self._types,
                          index=self._index(attr),
                          fallback=self._fallback)

    def _notify(self) -> None:
        types_str = self._types_str()
        match self._data, self._fallback, self._index():
            case GuardBase(), _, _:
                pass
            case _, _, []:
                pass
            case x , val, [*index] if x is Never:
                DefaultedReporter_m.add_defaulted(".".join(index), val, types_str)
            case val, _, [*index]:
                DefaultedReporter_m.add_defaulted(".".join(index), val, types_str)
            case val, flbck, index,:
                raise TypeError("Unexpected Values found: ", val, index, flbck)

    def _types_str(self) -> str:
        match self._types:
            case types.UnionType() as targ:
                types_str = repr(targ)
            case type(__name__=targ):
                types_str = targ
            case _ as targ:
                types_str = str(targ)

        return types_str

    def _match_type(self, val:TomlTypes) -> TomlTypes:
        if self._types != Any and not isinstance(val, self._types):
            types_str = self._types_str()
            index_str  = ".".join(self.__index + ['(' + types_str + ')'])
            msg = "TomlProxy Value doesn't match declared Type: "
            raise TypeError(msg, index_str, val, self._types)

        return val

    def _index(self, sub:str=None) -> list[str]:
        if sub is None:
            return self.__index[:]
        return self.__index[:] + [sub]



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import atexit#  for @atexit.register
import collections
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types as types_
import weakref
from copy import deepcopy
from time import sleep
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

from .._interface import TomlTypes

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class DefaultedReporter_m:
    """ A Mixin for reporting values that a failure proxy defaulted on. """

    _defaulted : ClassVar[set[str]] = set()

    @staticmethod
    def add_defaulted(index:str|list[str], val:TomlTypes, types:str="Any") -> None:
        match index, val:
            case list(), _:
                raise TypeError("Tried to Register a default value with a list index, use a str")
            case str(), bool():
                index_str = f"{index} = {str(val).lower()} # <{types}>"
            case str(), _:
                index_str = f"{index} = {repr(val)} # <{types}>"
            case [*xs], bool():
                index_path = ".".join(xs)
                index_str = f"{index_path} = {str(val).lower()} # <{types}>"
            case [*xs], _:
                index_path = ".".join(xs)
                index_str = f"{index_path} = {val} # <{types}>"
            case _, _:
                raise TypeError("Unexpected Values found: ", val, index)

        DefaultedReporter_m._defaulted.add(index_str)

    @staticmethod
    def report_defaulted() -> list[str]:
        """
        Report the index paths inject default values
        """
        return list(DefaultedReporter_m._defaulted)



#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types as types_
import weakref
from uuid import UUID, uuid1

##-- end builtin imports

from ..proxies.base import GuardProxy
from ..proxies.failure import GuardFailureProxy
from ..errors import GuardedAccessError

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class GuardProxyEntry_m:
    """ A Mixin to add to GuardBase.
    enables handling a number of conditions when accessing values in the underlying data.
    eg:
    tg.on_fail(2, int).a.value() # either get a.value, or 2. whichever returns has to be an int.
    """

    def on_fail(self, fallback:Any, types:Maybe[Any]=None, *, non_root=False) -> GuardFailureProxy:
        """
        use a fallback value in an access chain,
        eg: doot.config.on_fail("blah").this.doesnt.exist() -> "blah"

        *without* throwing a GuardedAccessError
        """
        index = self._index()
        if index != ["<root>"] and not non_root:
            raise GuardedAccessError("On Fail not declared at entry", index)

        return GuardFailureProxy(self, types=types, fallback=fallback)

    def first_of(self, fallback:Any, types:Maybe[Any]=None) -> GuardProxy:
        """
        get the first non-None value from a index path, even across arrays of tables
        so instead of: data.a.b.c[0].d
        just:          data.first_of().a.b.c.d()
        """
        raise NotImplementedError()

    def all_of(self, fallback:Any, types:Maybe[Any]=None) -> GuardProxy:
        raise NotImplementedError()

    def flatten_on(self, fallback:Any) -> GuardProxy:
        """
        combine all dicts at the call site to form a single dict
        """
        raise NotImplementedError()

    def match_on(self, **kwargs:tuple[str,Any]) -> GuardProxy:
        raise NotImplementedError()



#!/usr/bin/env python3
"""
Types that help add clarity

Provides a number of type aliases and shorthands.
Such as `Weak[T]` for a weakref, `Stack[T]`, `Queue[T]` etc for lists,
and `Maybe[T]`, `Result[T, E]`, `Either[L, R]`.

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import pathlib as pl
import types
from collections import deque
from collections.abc import (
    Callable,
    Generator,
    Hashable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    ValuesView,
)
from re import Match, Pattern
from typing import Annotated, Any, Never, Self, TypeGuard, final, Union, Final
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 3rd party imports
from packaging.specifiers import SpecifierSet
from packaging.version import Version

# ##-- end 3rd party imports

# ##-- Generated Exports
__all__ = ( # noqa: RUF022
# -- Types
"AbsPath", "Builder", "CHECKTYPE", "Char", "Ctor", "DateTime", "Decorator", "Depth",
"DictItems", "DictKeys", "DictVals", "E_", "Either", "Fifo", "FmtKey", "FmtSpec", "FmtStr",
"Frame", "Func", "Ident", "Lambda", "Lifo", "M_", "Maybe", "MaybeT", "Method", "Module", "Mut",
"NoMut", "Queue", "R_", "RelPath", "Result", "Rx", "RxMatch", "RxStr", "Seconds", "Stack",
"SubOf", "TimeDelta", "Traceback", "Url", "VList", "Vector", "VerSpecStr", "VerStr",
"Weak",

# -- Functions
"is_none",
)
# ##-- end Generated Exports

##-- strings
type VerStr                   = Annotated[str, Version] # A Version String
type VerSpecStr               = Annotated[str, SpecifierSet]
type Ident                    = Annotated[str, UUID] # Unique Identifier Strings
type FmtStr                   = Annotated[str, None] # Format Strings like 'blah {val} bloo'
type FmtSpec                  = Annotated[str, None] # Format and conversion parameters. eg: 'blah {val:<9!r}' would be ':<10!r'
type FmtKey                   = str # Names of Keys in a FmtStr
type RxStr                    = Annotated[str, Pattern]
type Char                     = Annotated[str, lambda x: len(x) == 1]
type Url                      = Annotated[str, "url"]

##-- end strings

##-- paths
type RelPath = Annotated[pl.Path, lambda x: not x.is_absolute()]
type AbsPath = Annotated[pl.Path, lambda x: x.is_absolute()]

##-- end paths

##-- regex
type Rx       = Pattern
type RxMatch  = Match

##-- end regex

##-- callables
type Ctor[T]            = type[T]
type Builder[T]         = Callable[[*Any], T]
type Func[**I, O]       = Callable[I, O]
type Method[**I, O]     = types.MethodType[I, O]
type Decorator[F:Func]  = Callable[[F], F]
type Lambda[**I, O]     = types.LambdaType[I, O]

##-- end callables

##-- containers
type Weak[T]    = ref[T]
type Stack[T]   = list[T]
type Fifo[T]    = list[T]
type Queue[T]   = deque[T]
type Lifo[T]    = list[T]
type Vector[T]  = list[T]

##-- end containers

##-- utils
type VList[T]                = T | list[T]
type Mut[T]                  = Annotated[T, "Mutable"]
type NoMut[T]                = Annotated[T, "Immutable"]

type Maybe[T]                = T | None
type MaybeT[*I]              = tuple[*I] | None
type Result[T, E:Exception]  = T | E
type Either[L, R]            = L | R
type SubOf[T]                = TypeGuard[T]

##-- end utils

##-- shorthands
type M_[T]                   = Maybe[T]
type R_[T, E:Exception]      = Result[T,E]
type E_[L, R]            = Either[L,R]

##-- end shorthands

##-- numbers
type Depth      = Annotated[int, lambda x: 0 <= x]
type Seconds    = Annotated[int, lambda x: 0 <= x]
type DateTime   = datetime.datetime
type TimeDelta  = datetime.timedelta

##-- end numbers

##-- dicts
type DictKeys   = KeysView
type DictItems  = ItemsView
type DictVals   = ValuesView

##-- end dicts

##-- tracebacks and frames
type Traceback = types.TracebackType
type Frame     = types.FrameType

##-- end tracebacks and frames

##-- misc
# the stdlib types.UnionType (int | float) is not typing.Union[int, float]
UnionTypes  : Final[types.UnionType]  = types.UnionType | type(Union[int,None])  # noqa: UP007
type Module                           = types.ModuleType
type CHECKTYPE                        = Maybe[type|types.GenericAlias|types.UnionType]

##-- end misc

##--| TypeGuards and TypeIs

def is_none(val:Maybe) -> TypeGuard[None]:
    return val is None



#!/usr/bin/env python3
"""

"""

# Imports:
##-- builtin imports
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
import tomllib
from uuid import UUID, uuid1

# ##-- end stdlib imports

##-- end builtin imports

# ##-- 1st party imports
from .._interface import TomlTypes

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard, TypeVar
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type T = TypeVar('T')

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


class TomlLoader_m:
    """ Mixin for loading toml files """

    @classmethod
    def read(cls:T, text:str) -> T:
        logging.debug("Reading ChainGuard for text")
        try:
            return cls(tomllib.loads(text))
        except Exception as err:
            raise IOError("ChainGuard Failed to Load: ", text, err.args) from err

    @classmethod
    def from_dict(cls, data:dict[str, TomlTypes]) -> Self:
        logging.debug("Making ChainGuard from dict")
        try:
            return cls(data)
        except Exception as err:
            raise IOError("ChainGuard Failed to Load: ", data, err.args) from err

    @classmethod
    def load(cls, *paths:str|pl.Path) -> Self:
        logging.debug("Creating ChainGuard for %s", paths)
        texts = []
        for path in paths:
            texts.append(pl.Path(path).read_text())
        else:
            try:
                return cls(tomllib.loads("\n".join(texts)))
            except tomllib.TOMLDecodeError as err:
                raise IOError("Failed to Load Toml", *err.args, paths) from err

    @classmethod
    def load_dir(cls, dirp:str|pl.Path) -> Self:
        logging.debug("Creating ChainGuard for directory: %s", str(dirp))
        try:
            texts = []
            for path in pl.Path(dirp).glob("*.toml"):
                texts.append(path.read_text())

            return cls(tomllib.loads("\n".join(texts)))
        except Exception as err:
            raise IOError("ChainGuard Failed to load Directory: ", dirp, err.args) from err



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011, ARG002, ANN001
from __future__ import annotations

import enum
import logging as logmod
import pathlib as pl
from typing import (Any, ClassVar, Generic, TypeAlias, TypeVar, cast, TYPE_CHECKING)
from collections.abc import Mapping
import warnings
import pytest

from jgdv.structs.strang import CodeReference

from jgdv.structs import dkey
from .. import _interface as API # noqa: N812
from .._interface import Key_p, DKeyMark_e
from ..processor import DKeyProcessor
from ..dkey import DKey
from .. import keys
from .. import special

if TYPE_CHECKING:
    from collections.abc import Generator

logging = logmod.root

class TestDKey_Mark:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic_mark(self):
        assert(isinstance(dkey.DKeyMark_e, enum.EnumMeta))

    def test_mark_aliases(self):
        assert("blah" not in dkey.DKeyMark_e)
        assert(dkey.DKeyMark_e.default() is Any)
        assert(dkey.DKeyMark_e.null() is False)
        assert(dkey.DKeyMark_e.multi() is list)
        assert(dkey.DKeyMark_e.indirect() is Mapping)

class TestDKey_Processor:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        obj = DKeyProcessor()
        match obj.pre_process(dkey.DKey, "test", implicit=True):
            case "test", dict() as inst_data, dict(), type() as ctor:
                assert(API.RAWKEY_ID in inst_data)
                assert(len(inst_data[API.RAWKEY_ID]) == 1)
                assert(isinstance(ctor, API.Key_p))
                assert(ctor.MarkOf(ctor) == Any), ctor
            case x:
                assert(False), x

    def test_basic_explicit(self):
        obj = DKeyProcessor()
        match obj.pre_process(dkey.DKey, "{test}"):
            case "test", dict() as inst_data, dict(), type() as ctor:
                assert(API.RAWKEY_ID in inst_data)
                assert(len(inst_data[API.RAWKEY_ID]) == 1)
                assert(isinstance(ctor, API.Key_p))
            case x:
                assert(False), x

    def test_multi_explicit(self):
        obj = DKeyProcessor()
        match obj.pre_process(dkey.DKey, "{test} mid {blah}"):
            case "{test} mid {blah}", dict() as inst_data, dict(), type() as ctor:
                assert(API.RAWKEY_ID in inst_data)
                assert(len(inst_data[API.RAWKEY_ID]) == 2)
                assert(isinstance(ctor, API.Key_p))
                assert(DKey.MarkOf(ctor) == list), ctor
            case x:
                assert(False), x

    def test_basic_explicit_with_format_params(self):
        obj = DKeyProcessor()
        match obj.pre_process(dkey.DKey, "{test:w}"):
            case "test", dict() as inst_data, dict(), type() as ctor:
                assert(API.RAWKEY_ID in inst_data)
                assert(len(inst_data[API.RAWKEY_ID]) == 1)
                assert(inst_data[API.RAWKEY_ID][0].key == "test")
                assert(inst_data[API.RAWKEY_ID][0].prefix == "")
                assert(inst_data[API.RAWKEY_ID][0].format == "w")
                assert(isinstance(ctor, API.Key_p))
            case x:
                assert(False), x

    def test_null_key(self):
        obj = DKeyProcessor()
        match obj.pre_process(dkey.DKey, "test"):
            case "test", dict() as inst_data, dict(), type() as ctor:
                assert(API.RAWKEY_ID in inst_data)
                assert(len(inst_data[API.RAWKEY_ID]) == 1)
                assert(isinstance(ctor, API.Key_p))
                assert(DKey.MarkOf(ctor) == False)
            case x:
                assert(False), x

class TestDKey_FormatParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_select_no_conversion(self):
        obj           = DKeyProcessor()
        default_ctor  = DKey._registry[DKey._default_k]
        match obj.pre_process(dkey.DKey, "test", implicit=True):
            case "test", dict(), dict(), type() as ctor:
                assert(ctor is default_ctor)
            case x:
                assert(False), x

    def test_select_strdkey(self):
        obj = DKeyProcessor()
        obj.register_convert_param(special.StrDKey, "s")
        match obj.pre_process(dkey.DKey, "test!s", implicit=True):
            case "test", dict(), dict(), type() as ctor:
                assert(ctor is special.StrDKey)
            case x:
                assert(False), x

    def test_select_indirect(self):
        obj = DKeyProcessor()
        obj.register_convert_param(keys.IndirectDKey, "I")
        match obj.pre_process(dkey.DKey, "test!I", implicit=True):
            case "test", dict(), dict(), type() as ctor:
                assert(ctor is keys.IndirectDKey)
            case x:
                assert(False), x

    def test_multikey_preprocess(self):
        obj = DKeyProcessor()
        obj.register_convert_param(keys.IndirectDKey, "I")
        obj.register_convert_param(special.StrDKey, "s")
        match obj.pre_process(dkey.DKey, "{test!I}} then {blah!s}"):
            case str(), dict(), dict(), type() as ctor:
                assert(ctor is keys.MultiDKey)
            case x:
                assert(False), x

    def test_multikey_build(self):
        obj = dkey.DKey("{test!I} then {blah!s}")
        assert(isinstance(obj, keys.MultiDKey))
        match obj.keys():
            case [keys.SingleDKey(), keys.SingleDKey()]:
                assert(True)
            case x:
                assert(False), x



#!/usr/bin/env python3
"""
The core implementation of the ChainGuard object,
which is then extended with mixins.
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from uuid import UUID, uuid1

# ##-- end stdlib imports

from .errors import GuardedAccessError
from .mixins.access_m import super_get, super_set
from . import _interface as API
from jgdv import Proto

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from collections.abc import ItemsView, KeysView, ValuesView
    from ._interface import TomlTypes
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

type InputData = dict[str, TomlTypes]

##--|

class GuardBase(dict):
    """
    Provides access to toml data (ChainGuard.load(apath))
    but as attributes (data.a.path.in.the.data)
    instead of key access (data['a']['path']['in']['the']['data'])

    while also providing typed, guarded access:
    data.on_fail("test", str | int).a.path.that.may.exist()

    while it can then report missing paths:
    data.report_defaulted() -> ['a.path.that.may.exist.<str|int>']
    """

    def __init__(self, data:Maybe[InputData]=None, *, index:Maybe[list[str]]=None, mutable:bool=False) -> None:
        super().__init__()
        super_set(self, "__table", data or {})
        super_set(self, "__index"   , (index or ["<root>"])[:])
        super_set(self, "__mutable" , mutable)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:{list(self.keys())}>"

    def __eq__(self, other:object) -> bool:
        match other:
            case GuardBase() as base:
                return  self._table() == base._table()
            case dict() as adict:
                return self._table() == adict
            case _:
                return False

    def __len__(self) -> int:
        return len(self._table())

    def __call__(self) -> TomlTypes:
        msg = "Don't call a ChainGuard, call a GuardProxy using methods like .on_fail"
        raise GuardedAccessError(msg)

    def __iter__(self) -> Iterator:
        return iter(getattr(self, "__table").keys())

    def __contains__(self, _key: object) -> bool:
        return _key in self.keys()

    def _index(self) -> list[str]:
        return super_get(self, "__index")[:]

    def _table(self) -> dict[str,TomlTypes]:
        return super_get(self, "__table")

    def keys(self) -> KeysView[str]: # type: ignore[override]
        table = super_get(self, "__table")
        return table.keys()

    def items(self) -> ItemsView[str, TomlTypes]: # type: ignore[override]
        match super_get(self, "__table"):
            case dict() as val:
                return val.items()
            case list() as val:
                return {self._index()[-1]: val}.items()
            case GuardBase() as val:
                return val.items()
            case x:
                msg = "Unknown table type"
                raise TypeError(msg, x)

    def values(self) -> list|ValuesView[TomlTypes]: # type: ignore[override]
        match super_get(self, "__table"):
            case dict() as val:
                return val.values()
            case list() as val:
                return val
            case _:
                raise TypeError()


    def update(self, *args) -> Never: # type: ignore[override]  # noqa: ANN002
        msg = "ChainGuards are immutable"
        raise NotImplementedError(msg)



#!/usr/bin/env python3
"""


"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
logging.disabled = True
##-- end logging

# Vars:

# Body:

class MetalordCore(type):
    """
    The Core of Metalord.

    By default classes are constructed using type().
    https://docs.python.org/3/reference/datamodel.html#metaclasses

    Execution order::

        - Metaclass Definition
        - Metaclass Creation

        - Class Definition:
            1) MRO resolution           : __mro_entries__(self, bases)
            2) metaclass selection      : (most dervied of (type | metaclasses))
            3) namespace preparation    : metacls.__prepare__(metacls, name, bases, **kwargs) -> dict
            4) class body execution     : exec(body, globals, namespace)
            5) class object creation    : metaclass.__new__(name, bases, namespace, **kwargs) -> classobj
            6) subclass init            : parent.__init_subclass__(cls, **kwargs)
            7) class object init        : metaclass.__init__(cls, na,e, bases, namespace, **kwargs)

        - Instance Creation::
            1) metaclass.__call__(cls, ...) -> instance
            1.b) cls.__new__(cls, ...) -> instance
            1.c) cls.__init__(instance, ...)

    """

    @classmethod
    def __prepare__(mcls:type, name:str, bases:tuple, **kwargs) -> dict:
        """ Class Definition preparation of namespace """
        logging.debug("Metalord Prepare: %s %s %s", name, bases, kwargs)
        return super().__prepare__(name, bases, **kwargs)

    def __new__(mcls:type, name:str, bases:tuple, namespace:dict, **kwargs) -> type:
        """ Class Definition Object creation
        Equivalent to type(name, bases, namespace)
        """
        logging.debug("Metalord new: %s %s %s %s %s", mcls, name, bases, namespace, kwargs)
        return super().__new__(mcls, name, bases, dict(namespace), **kwargs)

    def __init__(cls, name, bases:tuple, namespace:dict, **kwargs):
        """ Class Definition initialisation """
        logging.debug("Metalord init: %s %s %s %s %s", cls, name, bases, namespace, kwargs)
        super().__init__(name, bases, namespace, **kwargs)

    def __call__(cls, *args, **kwargs):
        """ Class Instance creation
        """
        logging.debug("Metalord Call: %s %s %s", cls, args, kwargs)
        obj = cls.__new__(cls, *args, **kwargs)
        obj.__init__(*args, **kwargs)
        return obj

    def __instancecheck__(self, instance) -> bool:
        logging.debug("Metalord instance check: %s %s", self, instance)
        return super().__instancecheck__(instance)

    def __subclasscheck__(self, subclass) -> bool:
        logging.debug("Metalord subclass check: %s %s", self, subclass)
        return super().__subclasscheck__(subclass)



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from jgdv.structs.strang import _interface as StrangAPI # noqa: N812
from jgdv.structs.strang._interface import Strang_p

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    import datetime
    import pathlib as pl
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Ident, Traceback
    from jgdv.structs.dkey._interface import Key_p

    type TimeDelta = datetime.timedelta
    type WordDesc = tuple[StrangAPI.StrangMarkAbstract_e, str]
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
LOC_SEP    : Final[str]   = "::>"
LOC_SUBSEP : Final[str]   = "/"

# Body:

class WildCard_e(StrangAPI.StrangMarkAbstract_e):
    """ Ways a path can have a wildcard. """
    glob       = "*"
    rec_glob   = "**"
    select     = "?"
    key        = "{"

class LocationMeta_e(StrangAPI.StrangMarkAbstract_e):
    """ Available metadata attachable to a location """

    location     = "location"
    directory    = "directory"
    file         = "file"

    abstract     = "abstract"
    artifact     = "artifact"
    clean        = "clean"
    earlycwd     = "earlycwd"
    protect      = "protect"
    expand       = "expand"
    remote       = "remote"
    partial      = "partial"

    # Aliases
    dir          = directory
    loc          = location

    @classmethod
    def default(cls) -> Maybe:
        return cls.location

##--|

LocationSections : Final[StrangAPI.Sections_d] = StrangAPI.Sections_d(
    StrangAPI.Sec_d("head", ".", LOC_SEP, str|LocationMeta_e, LocationMeta_e, False),  # noqa: FBT003
    StrangAPI.Sec_d("body", LOC_SUBSEP, None, str|WildCard_e, WildCard_e, True),  # noqa: FBT003
)
##--|

@runtime_checkable
class Location_p(Strang_p, Protocol):
    """ Something which describes a file system location,
    with a possible identifier, and metadata
    """
    Marks  : ClassVar[StrangAPI.StrangMarkAbstract_e]
    Wild   : ClassVar[StrangAPI.StrangMarkAbstract_e]

    def __lt__(self, other:TimeDelta|str|pl.Path|Location_p) -> bool: ...

    @property
    def keys(self) -> set[str]: ...

    @property
    def path(self) -> pl.Path: ...

    @property
    def body_parent(self) -> list[WordDesc]: ...

    @property
    def stem(self) -> Maybe[str|WordDesc]: ...

    @property
    def key(self) -> Maybe[str|Key_p]: ...

    def ext(self, *, last:bool=False) -> Maybe[str|WordDesc|tuple[str|WordDesc, ...]]: ...

    def check_wildcards(self, other:pl.Path|Location_p) -> bool: ...

    def is_concrete(self) -> bool: ...

@runtime_checkable
class Locator_p(Protocol):

    def __getattr__(self, key:str) -> Location_p: ...

    def __getitem__(self, val:str|pl.Path|Location_p|Key_p) -> pl.Path: ...

    def __contains__(self, key:str|pl.Path|Location_p|Key_p) -> bool: ...

    def __bool__(self) -> bool: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> Generator[str|Key_p]: ...

    def __call__(self, new_root:Maybe[pl.Path]=None) -> Locator_p: ...

    def __enter__(self) -> Locator_p: ...

    def __exit__(self, etype:Maybe[type[Exception]], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool: ...

    def clear(self) -> None: ...

    def update(self, extra:dict|Location_p|Locator_p, *, strict:bool=True) -> Self: ...



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from ..._interface import RawKey_d
from ..parser import DKeyParser

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
RAWKEY_ARGS : Final[str] = ["pre", "key", "format", "convert", "out"]
# Body:

class TestRawKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        match RawKey_d(prefix=""):
            case RawKey_d():
                assert(True)
            case x:
                assert(False), x


    def test_error_on_extra(self):
        with pytest.raises(AssertionError):
            RawKey_d(prefix="", non_val="blah")


    def test_error_on_no_prefix(self):
        with pytest.raises(KeyError):
            RawKey_d(key="blah")

    @pytest.mark.parametrize(RAWKEY_ARGS, [
        ("", "blah", "", "", "blah"),
        ("blah", "bloo", "", "", "bloo"),
        ("-- ", "awegah", "<50", "p", "awegah:<50!p"),
        ("-- ", "awegah_", "<50", "p", "awegah_:<50!p"),
    ])
    def test_joined(self, pre, key, format, convert, out):
        match RawKey_d(prefix=pre, format=format, key=key, convert=convert):
            case RawKey_d() as x if x.joined() == out:
                assert(True)
            case x:
                assert(False), (x, x.joined(), out)

    @pytest.mark.parametrize(RAWKEY_ARGS, [
        ("", "blah", "", "", "{}"),
        ("blah", "bloo", "", "", "blah{}"),
        ("-- ", "awegah", "<50", "p", "-- {}"),
        ("-- ", "awegah_", "", "p", "-- {}"),
    ])
    def test_anon(self, pre, key, format, convert, out):
        match RawKey_d(prefix=pre, key=key, format=format, convert=convert):
            case RawKey_d() as x if x.anon() == out:
                assert(True)
            case x:
                assert(False), (x, x.anon(), out)

    @pytest.mark.parametrize(RAWKEY_ARGS, [
        ("", "blah", "", "", "{blah}"),
        ("blah", "bloo", "", "", "{bloo}"),
        ("-- ", "awegah", "<50", "p", "{awegah}"),
        ("-- ", "awegah_", "<50", "p", "{awegah_}"),
    ])
    def test_wrapped(self, pre, key, format, convert, out):
        match RawKey_d(prefix=pre, key=key, format=format, convert=convert):
            case RawKey_d() as x if x.wrapped() == out:
                assert(True)
            case x:
                assert(False), (x, x.wrapped(), out)


    @pytest.mark.parametrize(RAWKEY_ARGS, [
        ("", "blah", "", "", "blah_"),
        ("blah", "bloo", "", "", "bloo_"),
        ("-- ", "awegah", "", "p", "awegah_"),
        ("", "aweg_", "", "", "aweg_"),
    ])
    def test_indirect(self, pre, key, format, convert, out):
        match RawKey_d(prefix=pre, key=key, format=format, convert=convert):
            case RawKey_d() as x if x.indirect() == out:
                assert(True)
            case x:
                assert(False), (x, x.indirect(), out)


    @pytest.mark.parametrize(RAWKEY_ARGS, [
        ("", "blah", "", "", "blah"),
        ("blah", "bloo", "", "", "bloo"),
        ("-- ", "awegah", "", "p", "awegah"),
        ("", "aweg_", "", "", "aweg"),
    ])
    def test_direct(self, pre, key, format, convert, out):
        match RawKey_d(prefix=pre, key=key, format=format, convert=convert):
            case RawKey_d() as x if x.direct() == out:
                assert(True)
            case x:
                assert(False), (x, x.direct(), out)

class TestParser:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_empty(self):
        match next(DKeyParser().parse("blah")):
            case RawKey_d() as x:
                assert(x.prefix == "blah")
                assert(True)
            case x:
                 assert(False), x

    def test_key(self):
        match next(DKeyParser().parse("{blah}")):
            case RawKey_d() as x:
                assert(x.prefix == "")
                assert(x.key == "blah")
                assert(True)
            case x:
                 assert(False), x

    def test_impicit(self):
        match next(DKeyParser().parse("blah", implicit=True)):
            case RawKey_d() as x:
                assert(x.prefix == "")
                assert(x.key == "blah")
                assert(True)
            case x:
                 assert(False), x

    def test_open_brace(self):
        match next(DKeyParser().parse("{blah")):
            case RawKey_d() as x:
                assert(x.key == None)
                assert(x.prefix == "{blah")
                assert(True)
            case x:
                 assert(False), x


    def test_close_brace(self):
        match next(DKeyParser().parse("blah}")):
            case RawKey_d() as x:
                assert(x.key == None)
                assert(x.prefix == "blah}")
                assert(True)
            case x:
                 assert(False), x



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011, PLR2004
# Imports
from __future__ import annotations

##-- stdlib imports
import logging as logmod
import pathlib as pl
from typing import (Any, ClassVar, Generic, TypeAlias,
                    TypeVar, cast)
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping, Sequence
from re import Match
from types import GenericAlias
import warnings

##-- end stdlib imports

import pytest
from .. import Subclasser
from ..annotate import SubAnnotate_m
from ..registrator import SubRegistry_m
from .._interface import AnnotationTarget

# Logging:
logging = logmod.root

# Global Vars:

class BasicEx(SubAnnotate_m):
    pass

class BasicSub(BasicEx):
    pass

class BasicTargeted(SubAnnotate_m, AnnotateTo="blah"):
    pass

##--|

class TestAnnotate_Registry:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_registry(self):

        class BasicReg(SubRegistry_m):
            pass

        class BasicSubReg(BasicReg[int]):
            pass

        class BasicSubOther(BasicReg[float]):
            pass

        assert(BasicReg[int] is BasicSubReg)
        assert(BasicReg[float] is not BasicReg[int])
        assert(int in BasicReg._registry)
        assert(BasicReg._registry is BasicSubReg._registry)




#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import builtins
import datetime
import enum
import functools as ftz
import importlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import typing
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Proto
from jgdv._abstract.pydantic_proto import ProtocolModelMeta
from jgdv.mixins.annotate import SubAlias_m
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from jgdv._abstract.protocols import Buildable_p
from jgdv.cli.errors import ArgParseError
from .. import _interface as API # noqa: N812
from .._interface import ParamStruct_p, ParamStruct_i

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Any, Literal
from collections.abc import Callable
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from jgdv import Rx
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class _SortGroups_e(enum.IntEnum):
    head      = 0
    by_prefix = 10
    by_pos    = 20
    last      = 99

##--|

class ParamProcessor:
    """ Parses a name into its component parts.

    eg: --blah= -> {prefix:--, name:blah, assign:None}
    """
    __slots__                 = ()

    name_re  : ClassVar[Rx]   = API.FULLNAME_RE
    end_sep  : ClassVar[str]  = API.END_SEP

   ##--| parsing

    def parse_name(self, name:str) -> Maybe[dict]:
        """
        Parse a string into a dict of the params:
        - name
        - prefix
        - separator
        """
        match self.name_re.match(name):
            case None:
                return None
            case re.Match() as matched:
                groups = matched.groupdict()

        result = {"name": matched['name'], "prefix":False}
        match groups:
            case {"pos": None|"", "prefix": None|""}:
                result['prefix'] = 99
            case {"pos": str() as x, "prefix":None}:
                result['prefix'] = int(x)
            case {"pos": None, "prefix":str() as x}:
                result['prefix'] = x

        match groups['assign']:
            case None:
                result['separator'] = False
            case str() as x:
                result['separator'] = x

        return result

    ##--| consuming

    def consume(self, obj:ParamStruct_i, args:list[str], *, offset:int=0) -> Maybe[tuple[dict, int]]:
        """
          Given a list of args, possibly add a value to the data.
          operates on both the args list
          return maybe(newdata, amount_consumed)

          handles:
          ["--arg=val"],
          ["-arg", "val"],
          ["val"],     (if positional=True)
          ["-arg"],    (if type=bool)
          ["-no-arg"], (if type=bool)
          """
        result  : Maybe[tuple[dict, int]]  = None
        consumed, remaining                = 0, args[offset:]
        ##--|
        logging.debug("Trying to consume: %s : %s", obj.name, remaining)
        try:
            match remaining:
                case []:
                    result = None
                case [x, *xs] if not self.matches_head(obj, x):
                    result = None
                case [*xs]:
                    key, value, consumed = self.next_value(obj, xs)
                    result = self.coerce_types(obj, key, value), consumed
                case _:
                    msg = "Tried to consume a bad type"
                    raise ArgParseError(msg, remaining)  # noqa: TRY301
        except ArgParseError as err:
            logging.debug("Parsing Failed: %s : %s (%s)", obj.name, args, err)
            return None
        else:
            return result

    def next_value(self, obj:ParamStruct_i, args:list) -> tuple[str, list, int]:
        match getattr(obj, "next_value", None):
            case None:
                pass
            case x:
                assert(callable(x))
                return x(args)

        if obj.type_ is bool:
            return obj.name, [args[0]], 1
        if obj.separator and obj.separator not in args[0]:
            return obj.name, [args[1]], 2

        key, *vals = self.split_assignment(obj, args[0])
        if key != obj.name:
            msg = "Assignment doesn't match"
            raise ArgParseError(msg, key, obj.name)
        return obj.name, [vals[0]], 1

    ##--| utils

    def coerce_types(self, obj:ParamStruct_i, key:str, value:list[str]) -> dict:
        """ coerce the parsed values to the expected type """
        x       : Any
        result  : dict[str, Any]  = {}
        match obj.type_, value:
            case x, y if x == type(y):
                result[key] = y
            case builtins.bool, [bool() as y]:
                result[key] = y
            case builtins.bool, y:
                result[key] = bool(y)
            case builtins.int, [*xs]: # sum a list of numbers
                result[key] = ftz.reduce(lambda x, y: x+int(y), xs, 0)
            case builtins.list, str() as x:
                result[key] = [x]
            case builtins.list, [*xs]:
                result[key] = list(xs)
            case builtins.set, [*xs]:
                result[key] = set(xs)
            case _, [x]: # unwrap things if they are a single value and not lists/sets
                result[key] = x
            case _, val:
                 result[key] = val

        return result

    def matches_head(self, obj:ParamStruct_i, val:str, *, silent:bool=True) -> bool:
        """ test to see if a cli argument matches this param

        Matchs {self.prefix}{self.name} if not an assignment
        Matches {self.prefix}{self.name}{separator} if an assignment
        """
        result : bool = False
        match getattr(obj, "matches_head", None):
            case None:
                key, *_ = self.split_assignment(obj, val)
                assert(isinstance(obj.key_strs, list))
                result = key in obj.key_strs and key.startswith(str(obj.prefix))
            case x:
                assert(callable(x))
                result = x(val)

        if result and not silent:
            logging.debug("Head Matches : %s : %s", obj.name, val)
        return result

    def match_on_end(self, val:str) -> bool:
        return val == self.end_sep

    def split_assignment(self, obj:ParamStruct_i, val:str) -> list[str]:
        if obj.separator:
            return val.split(obj.separator)
        return [val]

##--|

class _ParamClassMethods:

    @staticmethod
    def build_defaults( params:Iterable[ParamStruct_i]) -> dict:
        """ Given a list of params, create a mapping of {name} -> {default value} """
        result : dict = {}
        for p in params:
            assert(isinstance(p, ParamStruct_p)), repr(p)
            if p.name in result:
                msg = "Duplicate default key found"
                raise KeyError(msg, p, params)
            result.setdefault(*p.default_tuple)

        return result

    @staticmethod
    def check_insists(params:Iterable[ParamStruct_i], data:dict) -> None:
        missing = []
        for p in params:
            if p.insist and p.name not in data:
                missing.append(p.name)
        else:
            if bool(missing):
                msg = "Missing Required Params"
                raise ArgParseError(msg, missing)

    @classmethod
    def key_func(cls, x:ParamStruct_i) -> tuple:
        """ Sort Parameters

        -{prefix len} < name < int positional < positional < --help

        Positionals with explicit positions are sorted by that,
        With remaiing going after
        Help Always goes last.

        """
        match x.prefix:
            case _ if x.name == "help":
                return (_SortGroups_e.last, 99, x.prefix, x.name)
            case str():
                return (_SortGroups_e.by_prefix, len(x.prefix), x.prefix, x.name)
            case int() as p:
                return (_SortGroups_e.by_pos, p, x.prefix or 99, x.name)


class ParamSpec[*T](_ParamClassMethods, ParamStruct_i, SubAlias_m, fresh_registry=True):
    """ Declarative CLI Parameter Spec.

    | Declares the param name (turns into {prefix}{name})
    | The value will be parsed into a given {type}, and lifted to a list or set if necessary
    | If given, can have a {default} value.
    | {insist} will cause an error if it isn't parsed

    If {prefix} is a non-empty string, then its positional, and to parse it requires no -key.
    If {prefix} is an int, then the parameter has to be in the correct place in the given args.

    Can have a {desc} to help usage.
    Can be set using a short version of the name ({prefix}{name[0]}).
    If {implicit}, will not be listed when printing a param spec collection.

    """
    ##--| classvars
    _processor           : ClassVar[ParamProcessor]  = ParamProcessor()
    _accumulation_types  : ClassVar[list[Any]]       = [int, list, set]
    _pad                 : ClassVar[int]             = 15
    ##--| internal
    _short     : Maybe[str]
    _subtypes  : dict[type, type]

    ##--|

    ##--| dunders

    def __init__(self, **kwargs:Any) -> None:  # noqa: ANN401
        if bool({"prefix", "separator"} & kwargs.keys()):
            msg = "Don't specify prefix or separator explicitly, format the 'name' fully"
            raise KeyError(msg)
        super().__init__()
        match self._processor.parse_name(cast("str", kwargs.get("name"))):
            case dict() as ns:
                kwargs.update(ns)
            case _:
                pass

        self.name       = kwargs.pop("name")
        self.prefix     = kwargs.pop("prefix", API.DEFAULT_PREFIX)
        self.separator  = kwargs.pop("separator", False)
        self.type_      = kwargs.pop("type", None)
        self.insist     = kwargs.pop("insist", False)
        self.default    = kwargs.pop("default", None)
        self.desc       = kwargs.pop("desc", API.DEFAULT_DOC)
        self.count      = kwargs.pop("count", API.DEFAULT_COUNT)
        self.implicit   = kwargs.pop("implicit", False)

        self._short     =  None
        self._subtypes  =  {}

        self.validate_type()
        self.validate_default()


    def __repr__(self) -> str:
        match self.prefix:
            case str():
                return f"<{self.__class__.__name__}: {self.prefix}{self.name}>"
            case int():
                return f"<{self.__class__.__name__}: <{self.prefix}>{self.name}>"
            case x:
                raise TypeError(type(x))

    def __eq__(self, other:object) -> bool:
        match other:
            case API.ParamStruct_p() if type(self) is not type(other):
                    return False
            case API.ParamStruct_p(name=self.name, prefix=self.prefix, type_=self.type_, separator=self.separator): # type: ignore[misc]
                return True
            case _:
                return False

    ##--| validators

    def validate_type(self) -> None:  # noqa: PLR0912
        remap          : Any = None
        x              : Any
        override_type  : Maybe[type]  = getattr(self.__class__, self.__class__._annotate_to, None)
        match self.type_:
            case None | builtins.bool if override_type:
                self.type_ = override_type
            case typing.Any:
                self.type_ = None
                return
            case types.GenericAlias() as alias: # unwrap aliases
                self.type_ = alias.__origin__
                return
            case type() as x if isinstance(x, ParamStruct_p):
                msg = "Can't use a paramstruct as the type of a paramstruct"
                raise TypeError(msg, x)
            case type() as x if x is not bool:
                return
            case str() as x:
                remap = x
            case _:
                pass

        match self.cls_annotation():
            case _ if remap is not None:
                pass
            case None | ():
                remap = bool
            case [x]:
                remap = x
            case str() | type() | types.GenericAlias() as x:
                remap = x
            case x:
                raise TypeError(type(x))

        match API.TYPE_CONV_MAPPING.get(remap, remap):
            case types.GenericAlias() as x:
                self.type_ = x.__origin__
            case typing.Any:
                self.type_ = None
            case type() as x if not issubclass(x, ParamSpec):
                self.type_ = x
            case x:
                msg = "Bad Type for ParamSpec"
                raise TypeError(msg, x)

    def validate_default(self) -> None:
        match self.default:
            case "None":
                self.default = None
            case _:
                 pass

        match self.type_:
            case builtins.bool:
                self.default = self.default or False
            case builtins.int:
                self.default = self.default or 0
            case builtins.str:
                self.default = self.default or ""
            case builtins.list:
                self.default = self.default or list
            case builtins.set:
                self.default = self.default or set
            case types.GenericAlias() as x:
                self.default = self.default or x.__origin__
            case _:
                self.default = None

    ##--| properties

    @ftz.cached_property
    def short(self) -> str:
        return self._short or self.name[0]

    @ftz.cached_property
    def inverse(self) -> str:
        return f"no-{self.name}"

    @ftz.cached_property
    def repeatable(self) -> bool:
        return self.type_ in ParamSpec._accumulation_types

    @ftz.cached_property
    def key_str(self) -> str:
        """ Get how the param needs to be written in the cli.

        | eg: -test or --test
        """
        match self.prefix:
            case str():
                return f"{self.prefix}{self.name}"
            case _:
                return self.name

    @ftz.cached_property
    def short_key_str(self) -> Maybe[str]:
        match self.prefix:
            case str():
                return f"{self.prefix}{self.short}"
            case _:
                return None

    @ftz.cached_property
    def key_strs(self) -> list[str]:
        """ all available key-str variations """
        result = [self.key_str]
        match self.short_key_str:
            case str() as x:
                result.append(x)
            case None:
                pass

        match self.prefix:
            case str():
                inv = f"{self.prefix}{self.inverse}"
                result.append(inv)
            case _:
                pass

        return result

    @ftz.cached_property
    def default_value(self) -> Any:  # noqa: ANN401
        match self.default:
            case type() as ctor:
                return ctor()
            case x if callable(x):
                return x()
            case x:
                return x

    @ftz.cached_property
    def default_tuple(self) -> tuple[str, Any]:
        return self.name, self.default_value

    ##--| methods

    def consume(self, args:list[str], *, offset:int=0) -> Maybe[tuple[dict, int]]:
        return self._processor.consume(self, args, offset=offset)

    def help_str(self, *, force:bool=False) -> str:
        if self.implicit and not force:
            return ""

        match self.key_str:
            case None:
                parts = [f"[{self.name}]"]
            case str() as x:
                parts = [x]

        parts.append(" " * (self._pad - len(parts[0])))
        match self.type_:
            case type() if self.type_ is bool:
                parts.append(f"{'(bool)': <10}:")
            case str() if bool(self.default):
                parts.append(f"{'(str)': <10}:")
            case str():
                parts.append(f"{'(str)': <10}:")
            case _:
                pass

        parts.append(f"{self.desc:<30}")
        pad = " "*max(0, (85 - (len(parts)+sum(map(len, parts)))))
        match self.default:
            case None:
                pass
            case str():
                parts.append(f'{pad}: Defaults to: "{self.default}"')
            case _:
                parts.append(f"{pad}: Defaults to: {self.default}")

        return " ".join(parts)




## decorators.pyi -*- mode: python -*-
# Type Interface Specification
#
from __future__ import annotations

import enum
# ##-- types
# isort: off
import inspect
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv._abstract.types import Func, Method
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

# ##-- Generated Exports
__all__ = ( # noqa: RUF022
# -- Types
"Decorable", "Decorated", "Signature",
# -- Classes
"DForm_e", "Decorator_p",

)
# ##-- end Generated Exports


##--| Primary Types
type Signature          = inspect.Signature
type Decorable[**I, O]  = Callable[I, O]
type Decorated[**I, O]  = Callable[I, O]

##--| Val
WRAPPED             : Final[str]  = "__wrapped__"
ANNOTATIONS_PREFIX  : Final[str]  = "__JGDV__"
MARK_SUFFIX         : Final[str]  = "_mark"
DATA_SUFFIX         : Final[str]  = "_data"
ATTR_TARGET         : Final[str]  = "__annotations__"

class DForm_e(enum.Enum):
    """ This is necessary because you can't use Callable or MethodType
    in match statement
    """

    CLASS    = enum.auto()
    FUNC     = enum.auto()
    METHOD   = enum.auto()

##--|

class DecoratorHooks_p(Protocol):

    def _wrap_method_h[**In, Out](self, meth:Method[In,Out]) -> Decorated[In, Out]: ...

    def _wrap_fn_h[**In, Out](self, fn:Func[In, Out]) -> Decorated[In, Out]: ...

    def _wrap_class_h[T](self, cls:type[T]) -> Maybe[Decorated[[], T]]: ...

    def _validate_target_h(self, target:Decorable, form:DForm_e, args:Maybe[list]=None) -> None: ...

    def _validate_sig_h(self, sig:Signature, form:DForm_e, args:Maybe[list]=None) -> None: ...

    def _build_annotations_h(self, target:Decorable, current:list) -> list: ...

class DecoratorUtils_p(Protocol):

    def _decoration_logic(self, target:Decorable) -> Decorated: ...

    def _unwrap(self:Decorator_p, target:Decorated) -> Decorable: ...

    def _unwrapped_depth(self:Decorator_p, target:Decorated) -> int: ...

    def _build_wrapper(self:Decorator_p, form:DForm_e, target:Decorable) -> Maybe[Decorated]: ...

    def _apply_onto(self:Decorator_p, wrapper:Decorated, target:Decorable) -> Decorated: ...

    def _signature(self:Decorator_p, target:Decorable) -> Signature: ...

@runtime_checkable
class Decorator_p(DecoratorHooks_p, DecoratorUtils_p, Protocol):
    Form                 : ClassVar[type[DForm_e]]
    needs_args           : ClassVar[bool]

    _annotation_prefix   : str
    _data_key            : Maybe[str]
    _data_suffix         : str
    _mark_key            : Maybe[str]
    _mark_suffix         : str
    _wrapper_assignments : list[str]
    _wrapper_updates     : list[str]

    def __call__[**I, O](self, target:Decorable[I, O]) -> Decorated[I, O]: ...

    def data_key(self) -> str: ...

    def mark_key(self) -> str: ...

    def dec_name(self) -> str: ...

    def apply_mark(self, *args:Decorable) -> None: ...

    def annotate_decorable(self, target:Decorable) -> list: ...

    def is_marked(self, target:Decorable) -> bool: ...

    def is_annotated(self, target:Decorable) -> bool: ...

    def get_annotations(self, target:Decorable) -> list[str]: ...

##--| Interface



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast, Self, Final)
import warnings

import pytest

from jgdv.structs.locator import JGDVLocator
from ... import DKey, Key_p
from ..path_key import PathDKey

logging = logmod.root

IMP_KEY_BASES               : Final[list[str]]           = ["bob", "bill", "blah", "other", "23boo", "aweg2531", "awe_weg", "aweg-weji-joi"]
EXP_KEY_BASES               : Final[list[str]]           = [f"{{{x}}}" for x in IMP_KEY_BASES]
EXP_P_KEY_BASES             : Final[list[str]]           = ["{bob:wd}", "{bill:w}", "{blah:wi}", "{other:i}"]
PATH_KEYS                   : Final[list[str]]           = ["{bob}/{bill}", "{blah}/{bloo}", "{blah}/{bloo}"]
MUTI_KEYS                   : Final[list[str]]           = ["{bob}_{bill}", "{blah} <> {bloo}", "! {blah}! {bloo}!"]
IMP_IND_KEYS                : Final[list[str]]           = ["bob_", "bill_", "blah_", "other_"]
EXP_IND_KEYS                : Final[list[str]]           = [f"{{{x}}}" for x in IMP_IND_KEYS]

VALID_KEYS                                           = IMP_KEY_BASES + EXP_KEY_BASES + EXP_P_KEY_BASES + IMP_IND_KEYS + EXP_IND_KEYS
VALID_MULTI_KEYS                                     = PATH_KEYS + MUTI_KEYS

class TestPathKey:

    def test_mark(self):
        assert(DKey.MarkOf(PathDKey) is pl.Path)

    def test_expansion(self):
        key = DKey[pl.Path]("test", implicit=True)
        match key.expand({"test":"blah"}):
            case pl.Path() as x:
                assert(x == pl.Path.cwd() / "blah")
                assert(True)
            case x:
                 assert(False), x


    @pytest.mark.xfail
    def test_cwd_expansion(self):
        cwd = pl.Path.cwd()
        key = DKey[pl.Path](".")
        match key.expand({"test":"blah"}):
            case pl.Path() as x:
                assert(x == cwd)
            case x:
                 assert(False), x


    def test_expansion_fail(self):
        key = DKey[pl.Path]("test", implicit=True)
        match key.expand():
            case None:
                assert(True)
            case x:
                 assert(False), x


    def test_loc_expansion(self):
        locs = JGDVLocator(root=pl.Path.cwd())
        locs.update({"test":"blah"})
        key = DKey[pl.Path]("test", implicit=True)
        match key.expand(locs):
            case pl.Path() as x:
                assert(x == pl.Path.cwd() / "blah")
                assert(True)
            case x:
                 assert(False), x


    def test_loc_expansion_miss(self):
        locs = JGDVLocator(root=pl.Path.cwd())
        key = DKey("test", force=PathDKey, implicit=True)
        match key.expand(locs):
            case None:
                assert(True)
            case x:
                 assert(False), x


    @pytest.mark.skip
    def test_todo(self):
        pass



"""
A module is for file format readers and writers.

BookmarkCollection simplifies managing my bookmarks
TagFile is for tracking the tags I use in bibtex/bookmarks
SubFile is for correcting mispelled tags and merging tags en masse.

"""



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import builtins
import datetime
import enum
import functools as ftz
import importlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import typing
import weakref
from dataclasses import InitVar, dataclass, field
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Maybe
from jgdv.mixins.annotate.annotate import SubAnnotate_m
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from jgdv import Proto
from jgdv.cli.errors import ArgParseError
from .. import _interface as API # noqa: N812
from .param_spec import ParamSpec

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Any
from collections.abc import Callable
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Any, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|
from .._interface import ParamStruct_p
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class ToggleParam(ParamSpec, annotation=bool, default=True):
    """ A bool of -param or -not-param """

    desc : str = "A Toggle"

    def __init__(self, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        super().__init__(*args, **kwargs)
        if self.type_ is not bool:
            msg = "Toggle Params can only be boolean"
            raise TypeError(msg, self.name)

    def _toggle(self) -> Literal[True]:
        return True

    def next_value(self, args:list) -> tuple[str, list, int]:
        head, *_ = args
        if self.inverse in head:
            value = self.default_value
        else:
            value = not self.default_value

        return self.name, [value], 1

class KeyParam(ParamSpec[str]):
    """ a param that is specified by a prefix key

    eg: -key val
    """
    desc  : str = "A Key"

    def _keyval(self) -> Literal[True]:
        return True

    def matches_head(self, val:str) -> bool:
        return val in self.key_strs

    def next_value(self, args:list) -> tuple[str, list, int]:
        """ get the value for a -key val """
        logging.debug("Getting Key/Value: %s : %s", self.name, args)
        match args:
            case [x, y, *_] if self.matches_head(x):
                return self.name, [y], 2
            case _:
                msg = "Failed to parse key"
                raise ArgParseError(msg)

class AssignParam(ParamSpec):
    """ a joined --key=val param """

    desc : str = "An Assignment Param"

    def __init__(self, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        kwargs.setdefault("type", str)
        super().__init__(*args, **kwargs)
        if self.type_ is bool:
            msg = "A boolean assignment param is pointless, use a toggle"
            raise TypeError(msg, self.name)
        match self.prefix:
            case str() as x if bool(x):
                pass
            case x:
                msg = "Bad prefix value for an assignment param"
                raise ValueError(msg, x)
        match self.separator:
            case str() as x if bool(x):
                pass
            case x:
                msg = "Bad separator value for an assignment param"
                raise ValueError(msg, x)

    def _assignment(self) -> Literal[True]:
        return True

    def next_value(self, args:list) -> tuple[str, list, int]:
        """ get the value for a --key=val """
        logging.debug("Getting Key Assignment: %s : %s", self.name, args)
        if self.separator not in args[0]:
            msg = "Assignment param has no assignment"
            raise ArgParseError(msg, self.separator, args[0])
        key,val = self._processor.split_assignment(self, args[0])
        return self.name, [val], 1

class PositionalParam(ParamSpec):
    """ A param that is specified by its position in the arg list

    Positional Params are formatted as <int>name.
    eg: <2>blah

    The integer is the *relative* sort of the parameter.
    As the full parameter list can be accumulated at runtime,
    ParamSpec sorts them ready for use.
    See ParamSpec.key_func.

    Suffice to say, at specification time: <2>blah <50>aweg <-2>qqqq
    Results in a run time positional param list of: qqqq, blah, aweg

    """

    desc : str = "A Positional Param"

    def _positional(self) -> Literal[True]:
        return True

    @ftz.cached_property
    def key_str(self) -> str:
        return self.name

    def matches_head(self, val:str) -> bool:  # noqa: ARG002
        return True

    def next_value(self, args:list) -> tuple[str, list, int]:
        match self.count:
            case API.DEFAULT_COUNT:
                return self.name, [args[0]], 1
            case API.UNRESTRICTED_COUNT if self._processor.end_sep in args:
                idx     = args.index(self._processor.end_sep)
                claimed = args[max(idx, len(args))]
                return self.name, claimed, len(claimed)
            case API.UNRESTRICTED_COUNT:
                return self.name, args[:], len(args)
            case int() as x if x < len(args):
                return self.name, args[:x], x
            case x:
                msg = "Bad positional count"
                raise ArgParseError(msg, x)


class LiteralParam(ToggleParam):
    """
    Match on a Literal Parameter.
    For command/subcmd names etc
    """
    desc   : str = "A Literal"

    def matches_head(self, val:str) -> bool:
        """ test to see if a cli argument matches this param """
        match val:
            case x if x == self.key_str:
                return True
            case _:
                return False



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011, ERA001, F841

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
from collections import ChainMap
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import identity_fn
# ##-- end 1st party imports

from .._interface import ExpInst_d, SourceChain_d
from ... import DKey, IndirectDKey, NonDKey
from ..expander import DKeyExpander

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from collections.abc import Mapping

if TYPE_CHECKING:
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestExpInst_d:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        obj = ExpInst_d(value="blah", fallback="bloo")
        assert(obj.value == "blah")
        assert(obj.rec == -1)
        assert(obj.fallback == "bloo")

    def test_no_val_errors(self):
        with pytest.raises(KeyError):
            ExpInst_d(fallback="bloo")

    def test_match(self):
        match ExpInst_d(value="blah", fallback="bloo"):
            case ExpInst_d(value="blah"):
                assert(True)
            case x:
                assert(False), x

    def test_match_fail(self):
        match ExpInst_d(value="bloo", fallback="bloo"):
            case ExpInst_d(rec=True):
                assert(False)
            case _:
                assert(True)

class TestSourceChain_d:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        match SourceChain_d():
            case SourceChain_d():
                assert(True)
            case x:
                assert(False), x

    def test_with_base(self):
        match SourceChain_d({"a":2, "b":3}):
            case SourceChain_d():
                assert(True)
            case x:
                assert(False), x

    def test_with_specstruct_base(self):

        class SimpleSpecStruct:

            @property
            def params(self) -> dict:
                return {"blah":"aweg"}

            @property
            def args(self) -> list:
                return []

            @property
            def kwargs(self) -> dict:
                return {}

        match SourceChain_d(SimpleSpecStruct()):
            case SourceChain_d() as obj:
                assert(obj.get("blah") == "aweg")
            case x:
                assert(False), x

    def test_with_multi_base(self):
        match SourceChain_d({"a":2, "b":3}, {"blah":"bloo"}):
            case SourceChain_d():
                assert(True)
            case x:
                assert(False), x

    def test_extend(self):
        obj = SourceChain_d({"a":2, "b":3}, {"blah":"bloo"})
        match obj.extend({"blee":"aweg"}):
            case SourceChain_d() as extended:
                assert(obj is extended)
                assert(obj.get("blee") == "aweg")
            case x:
                assert(False), x

    def test_simple_get(self):
        obj = SourceChain_d({"a":2, "b":3})
        match obj.get("b"):
            case 3:
                assert(True)
            case x:
                assert(False), x

    def test_simple_get_with_fallback(self):
        obj = SourceChain_d({"a":2, "b":3})
        match obj.get("d", 10):
            case 10:
                assert(True)
            case x:
                assert(False), x

    def test_multi_base_get(self):
        obj = SourceChain_d({"a":2, "b":3}, {"blah":"bloo"})
        match obj.get("blah"):
            case "bloo":
                assert(True)
            case x:
                assert(False), x

    def test_lookup(self):
        obj   = SourceChain_d({"a":2, "b":3}, {"blah":"bloo"})
        inst  = ExpInst_d(value="blah")
        match obj.lookup([inst]):
            case ExpInst_d(value="bloo"):
                assert(True)
            case x:
                assert(False), x

class TestExpander:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match DKeyExpander():
            case DKeyExpander() as x:
                assert(x._ctor is None)
            case x:
                assert(False), x

    def test_set_ctor(self):
        obj = DKeyExpander()
        assert(obj._ctor is None)
        obj.set_ctor(DKey)
        assert(obj._ctor is DKey)

    def test_extra_sources(self):
        obj = DKeyExpander()

        class SimpleDKey(DKey):
            __slots__ = ()

            def exp_extra_sources_h(self, sources):
                sources.extend([1,2,3,4])
                return sources

        key = DKey("blah", force=SimpleDKey)
        assert(isinstance(key, SimpleDKey))
        match obj.extra_sources([], source=key):
            case SourceChain_d() as x:
                assert(x.sources[0] == [1,2,3,4])
            case x:
                assert(False), x
        match key.exp_extra_sources_h(SourceChain_d()):
            case SourceChain_d() as x:
                assert(x.sources[0] == [1,2,3,4])
            case x:
                assert(False), x

    def test_generate_alternatives(self):
        obj = DKeyExpander()
        key = DKey("{simple}")
        match obj.generate_alternatives([], {}, source=key):
            case [[ExpInst_d(), ExpInst_d()]]:
                assert(True)
            case x:
                assert(False), x

    def test_do_lookup(self):
        obj      = DKeyExpander()
        obj.set_ctor(DKey)
        key      = DKey("{simple}")
        insts    = [[ExpInst_d(value="simple")]]
        sources  = SourceChain_d({"simple":"blah"})
        match obj.do_lookup(insts, sources, {}, source=key):
            case [ExpInst_d(value="blah")]:
                assert(True)
            case x:
                assert(False), x

    def test_configure_recursion_default(self):
        obj      = DKeyExpander()
        key      = DKey("{simple}")
        insts    = [ExpInst_d(value="simple")]
        sources  = [{"simple":"blah"}]
        match obj.configure_recursion(insts, sources, {}, source=key):
            case list() as result:
                assert(result is insts)
            case x:
                assert(False), x

    def test_do_recursion(self):
        obj      = DKeyExpander()
        obj.set_ctor(DKey)
        key      = DKey("{simple}")
        insts    = [ExpInst_d(value=key, rec=1)]
        sources  = ChainMap({"simple":"blah"})
        match obj.do_recursion(insts, sources, {}, source=key):
            case [ExpInst_d(value="blah")]:
                assert(True)
            case x:
                assert(False), x

    def test_flatten(self):
        obj      = DKeyExpander()
        obj.set_ctor(DKey)
        key      = DKey("{simple}")
        insts    = [ExpInst_d(value=key, rec=1)]
        sources  = [{"simple":"blah"}]
        match obj.flatten(insts, {}, source=key):
            case ExpInst_d() as x if x is insts[0]:
                assert(True)
            case x:
                assert(False), x

    def test_coerce_result_no_op(self):
        obj      = DKeyExpander()
        obj.set_ctor(DKey)
        key      = DKey("{simple}")
        inst     = ExpInst_d(value=key, rec=1)
        sources  = [{"simple":"blah"}]
        match obj.coerce_result(inst, {}, source=key):
            case ExpInst_d(value=str(), literal=True) as x:
                assert(True)
            case x:
                assert(False), x

    def test_coerce_result_simple(self):
        obj      = DKeyExpander()
        obj.set_ctor(DKey)
        key      = DKey("{simple}")
        key.data.expansion_type = pl.Path
        inst     = ExpInst_d(value=key, rec=1)
        sources  = [{"simple":"blah"}]
        match obj.coerce_result(inst, {}, source=key):
            case ExpInst_d(value=pl.Path(), literal=True) as x:
                assert(True)
            case x:
                assert(False), x

    @pytest.mark.skip
    def test_finalise(self):
        pass

    @pytest.mark.skip
    def test_check_result(self):
        pass

class TestExpansion:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        """ {test} -> blah """
        obj = DKey("test", implicit=True)
        state = {"test": "blah"}
        match obj.expand(state):
            case "blah":
                assert(True)
            case x:
                assert(False), x

    def test_basic_fail(self):
        """ {aweg} -> None """
        obj = DKey("aweg", implicit=True)
        state = {"test": "blah"}
        match obj.expand(state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_nonkey_expansion(self):
        """ aweg -> aweg """
        obj = DKey("aweg")
        state = {"test": "blah"}
        match obj.expand(state):
            case "aweg":
                assert(True)
            case x:
                assert(False), x

    def test_simple_recursive(self):
        """
        {test} -> {blah} -> bloo
        """
        obj = DKey("test", implicit=True)
        state = {"test": "{blah}", "blah": "bloo"}
        match obj.expand(state):
            case "bloo":
                assert(True)
            case x:
                assert(False), x

    def test_double_recursive(self):
        """
        {test} -> {blah}
        {blah} -> {aweg}/{bloo}
        {aweg}/{bloo} -> qqqq/{aweg}
        qqqq/{aweg} -> qqqq/qqqq
        """
        obj = DKey("test", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}/{bloo}", "aweg":"qqqq", "bloo":"{aweg}"}
        match obj.expand(state):
            case "qqqq/qqqq":
                assert(True)
            case x:
                assert(False), x

    def test_coerce_type(self):
        """ test -> str(blah) -> pl.Path(blah) """
        obj = DKey("test", implicit=True, ctor=pl.Path)
        state = {"test": "blah"}
        match obj.expand(state):
            case pl.Path():
                assert(True)
            case x:
                assert(False), x

    def test_check_type(self):
        """ {test} -> pl.Path(blah) """
        obj = DKey("test", implicit=True, ctor=pl.Path)
        state = {"test": pl.Path("blah")}
        match obj.expand(state):
            case pl.Path():
                assert(True)
            case x:
                assert(False), x

    def test_expansion_cascade(self):
        """
        {test} -1-> {blah},
        {test} -2-> {aweg}
        {test} -3-> qqqq
        """
        obj = DKey("test", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(obj.expand(state, limit=1) == "blah")
        assert(obj.expand(state, limit=2) == "aweg")
        match obj.expand(state, limit=3):
            case NonDKey() as x:
                assert(x == "qqqq")
            case x:
                assert(False), x

    def test_expansion_limit_format_param(self):
        """
        {test:e1} -> state[test:{blah}, blah:{aweg}, aweg:qqqq] -> {bloo}
        """
        obj = DKey("test:e1", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(obj.expand(state) == "blah")

    def test_expansion_limit_format_param_two(self):
        """
        {test:e2} -> state[test:{blah}, blah:{aweg}, aweg:qqqq] -> {aweg}
        """
        obj = DKey("test:e2", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(obj.expand(state) == "aweg")
        assert(isinstance(obj.expand(state), DKey))

    @pytest.mark.skip("TODO")
    def test_additional_sources_recurse(self):
        """ see doot test_dkey.TestDKeyExpansion.test_indirect_wrapped_expansion
        """
        assert(False)

    def test_keep_original_type_on_expansion(self):
        """ {test} -> state[test:[1,2,3]] -> [1,2,3] """
        obj = DKey("test", implicit=True)
        state = {"test": [1,2,3]}
        match obj.expand(state):
            case list():
                assert(True)
            case x:
                assert(False), type(x)

class TestIndirection:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_hit(self):
        """
        {key} -> state[key:value] -> value
        """
        obj = DKey("test", implicit=True)
        state = {"test": "blah"}
        match obj.expand(state):
            case "blah":
                assert(True)
            case x:
                assert(False), x

    def test_hit_ignores_indirect(self):
        """
        {key} -> state[key:value, key_:val2] -> value
        """
        obj = DKey("test", implicit=True)
        state = {"test": "blah", "test_":"aweg"}
        match obj.expand(state):
            case "blah":
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss(self):
        """
        {key} -> state[] -> None
        """
        obj = DKey("test", implicit=True)
        state = {}
        match obj.expand(state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_with_call_fallback(self):
        """
        {key} -> state[] -> 25
        """
        obj = DKey("test", implicit=True)
        state = {}
        match obj.expand(state, fallback=25):
            case 25:
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_with_ctor_fallback(self):
        """
        {key} -> state[] -> 25
        """
        obj = DKey("test", fallback=25, implicit=True)
        state = {}
        match obj.expand(state):
            case 25:
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_prefer_call_fallback_over_ctor(self):
        """
        {key} -> state[] -> 25
        """
        obj = DKey("test", fallback=10, implicit=True)
        state = {}
        match obj.expand(state, fallback=25):
            case 25:
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_indirect(self):
        """
        {key_} -> state[] -> {key_}
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) == Mapping)
        state = {}
        match obj.expand(state):
            case DKey() as k if k == "test_":
                assert(DKey.MarkOf(k) == Mapping)
                assert(True)
            case x:
                assert(False), x

    def test_soft_miss(self):
        """
        {key} -> state[key_:blah] -> {blah}
        """
        target  = DKey("blah", implicit=True)
        obj     = DKey("test", implicit=True)
        state   = {"test_": "blah"}
        match obj.expand(state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_soft_hit_direct(self):
        """
        {key_} -> state[key:value] -> value
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) is Mapping)
        state = {"test": "blah"}
        match obj.expand(state):
            case "blah":
                assert(True)
            case x:
                assert(False), x

    def test_soft_hit_indirect(self):
        """
        {key_} -> state[key_:key2, key2:value] -> {value}
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) is Mapping)
        state = {"test_": "blah", "blah":"bloo"}
        match obj.expand(state):
            case "bloo":
                assert(True)
            case x:
                assert(False), x

    def test_indirect_prefers_indirect_over_direct(self):
        """
        {key_} -> state[key_:value, key:val2] -> {value}
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) is Mapping)
        state = {"test_": "blah", "test": "aweg", "blah":"bloo"}
        match obj.expand(state):
            case "bloo":
                assert(True)
            case x:
                assert(False), x

class TestMultiExpansion:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        obj = DKey("{test} {test}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match obj.expand(state):
            case "blah blah":
                assert(True)
            case x:
                assert(False), x

    def test_coerce_to_path(self):
        obj = DKey("{test}/{test}", ctor=pl.Path)
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match obj.expand(state):
            case pl.Path():
                assert(True)
            case x:
                assert(False), x

    def test_coerce_subkey(self):
        obj = DKey("{test!p}/{test}")
        assert(DKey.MarkOf(obj) is list)
        assert(obj.keys()[0].data.convert == "p")
        state = {"test": "blah"}
        match obj.expand(state):
            case str() as x:
                assert(x == str(pl.Path.cwd() / "blah/blah"))
                assert(True)
            case x:
                assert(False), x

    def test_coerce_multi(self):
        obj = DKey("{test!p} : {test!p}")
        assert(DKey.MarkOf(obj) is list)
        assert(obj.keys()[0].data.convert == "p")
        state = {"test": "blah"}
        match obj.expand(state):
            case str() as x:
                assert(x == "".join([str(pl.Path.cwd() / "blah"),
                                    " : ",
                                    str(pl.Path.cwd() / "blah")]))
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_subkey(self):
        """ {key}/{key2} -> state[key:value} -> None """
        obj = DKey("{test}/{aweg}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match obj.expand(state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_soft_miss_subkey(self):
        obj = DKey("{test}/{aweg}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah", "aweg_":"test"}
        match obj.expand(state):
            case "blah/blah":
                assert(True)
            case x:
                assert(False), x

    def test_indirect_subkey(self):
        obj = DKey("{test}/{aweg_}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah", "aweg_":"test"}
        match obj.expand(state):
            case "blah/blah":
                assert(True)
            case x:
                assert(False), x

    def test_indirect_key_subkey(self):
        obj = DKey("{test}/{aweg_}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah", "aweg":"test"}
        match obj.expand(state):
            case "blah/test":
                assert(True)
            case x:
                assert(False), x

    def test_indirect_miss_subkey(self):
        obj = DKey("{test}/{aweg_}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match obj.expand(state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_multikey_of_one(self):
        obj = DKey[list]("{test}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "{blah}", "blah": "blah/{aweg_}"}
        match obj.expand(state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_multikey_recursion(self):
        obj = DKey[list]("{test}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "{test}", "blah": "blah/{aweg_}"}
        match obj.expand(state, limit=10):
            case "{test}":
                assert(True)
            case x:
                assert(False), x

    def test_multikey_recursion_limit(self):
        """
        {test:e2} -> state[test:{blah}, blah:{aweg}, aweg:qqqq] -> {aweg}
        """
        obj = DKey("{test:e1} : {test:e2}")
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(obj.expand(state) == "{blah} : {aweg}")

class TestCoercion:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_coerce_param_path(self):
        obj = DKey("{test!p}")
        state = {"test": "blah"}
        assert(obj.data.convert == "p")
        match obj.expand(state):
            case pl.Path():
                assert(True)
            case x:
                assert(False), x

    def test_coerce_param_int(self):
        obj = DKey("{test!i}")
        state = {"test": "25"}
        assert(obj.data.convert == "i")
        match obj.expand(state):
            case 25:
                assert(True)
            case x:
                assert(False), x

    def test_coerce_param_fail(self):
        obj = DKey("{test!i}")
        state = {"test": "blah"}
        assert(obj.data.convert == "i")
        with pytest.raises(ValueError):
            obj.expand(state)

class TestFallbacks:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic_fallback(self):
        key = DKey("blah", implicit=True, fallback="aweg")
        match key():
            case "aweg":
                assert(True)
            case x:
                 assert(False), x

    def test_fallback_typecheck(self):
        key = DKey("blah", implicit=True, fallback="aweg", check=str)
        match key():
            case "aweg":
                assert(True)
            case x:
                 assert(False), x

    @pytest.mark.parametrize("ctor", [list, dict, set])
    def test_fallback_type_factory(self, ctor):
        key = DKey("blah", implicit=True, fallback=ctor)
        match key():
            case list()|dict()|set():
                assert(True)
            case x:
                 assert(False), x



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011, PLR2004
# Imports
from __future__ import annotations

##-- stdlib imports
import logging as logmod
import pathlib as pl
from typing import (Any, ClassVar, Generic, TypeAlias,
                    TypeVar, cast)
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping, Sequence
from re import Match
from types import GenericAlias
import warnings

##-- end stdlib imports

import pytest
from .. import Subclasser
from .._interface import AnnotationTarget

# Logging:
logging = logmod.root

# Global Vars:

class BasicEx:
    pass

class BasicSub(BasicEx):
    pass

##--|

class TestSubClasser_Decoration:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_decorate_name_noop(self):
        match Subclasser.decorate_name(BasicEx):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_decorate_name_with_extras(self):
        target = f"{BasicEx.__name__}<+test>"
        match Subclasser.decorate_name(BasicEx, "test"):
            case str() as x if x == target:
                assert(True)
            case x:
                 assert(False), x

    def test_decorate_name_idempotent(self):
        target = f"{BasicEx.__name__}<+test>"
        r1 = Subclasser.decorate_name(BasicEx, "test")
        r2 = Subclasser.decorate_name(r1, "test")
        assert(target == r1 == r2)

    def test_redecorate_name_with_extras(self):
        target = f"{BasicEx.__name__}<+test>"
        curr = BasicEx.__name__
        for _ in range(10):
            curr = Subclasser.decorate_name(curr, "test")
            assert(curr == target)

    def test_multi_decorate(self):
        target = f"{BasicEx.__name__}<+blah+test>"
        curr = BasicEx.__name__
        for _ in range(10):
            curr = Subclasser.decorate_name(curr, "test", "blah")
            assert(curr == target)

    def test_decorate_param(self):
        target = f"{BasicEx.__name__}[bool]"
        curr = BasicEx.__name__
        for _ in range(10):
            curr = Subclasser.decorate_name(curr, params="bool")
            assert(curr == target)

    def test_decorate_name_override_param(self):
        target = f"{BasicEx.__name__}[int]"
        curr = Subclasser.decorate_name(BasicEx.__name__, params="bool")
        over = Subclasser.decorate_name(curr, params="int")
        assert(over == target)

    def test_decorate_extras_and_params(self):
        target = f"{BasicEx.__name__}<+test>[int]"
        curr = Subclasser.decorate_name(BasicEx.__name__, "test", params="int")
        assert(curr == target)

class TestSubclasser_Making:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic_subclass(self) -> None:
        builder        = Subclasser()
        target  : str  = "TestCls"

        class Basic:
            val : int
            pass

        match builder.make_subclass(target, Basic):
            case type() as x:
                assert(issubclass(x, Basic))
                assert(x.__name__ == target)
                assert(Basic in x.mro())
                inst = x()
                assert(isinstance(inst, x))
                assert(isinstance(inst, Basic))
            case x:
                assert(False), x

    def test_basic_preserves_slots(self) -> None:
        builder        = Subclasser()
        target  : str  = "TestCls"

        class Basic:
            __slots__ = ("val",)

        class ExSub(Basic):
            """ An example subclass the usual way """
            __slots__ = ()

        ex_inst = ExSub()
        assert(not hasattr(ex_inst, "__dict__"))

        match builder.make_subclass(target, Basic):
            case type() as x:
                assert(issubclass(x, Basic))
                assert(x.__name__ == target)
                assert(Basic in x.mro())
                inst = x()
                assert(isinstance(inst, x))
                assert(isinstance(inst, Basic))
                assert(not hasattr(inst, "__dict__"))
            case x:
                assert(False), x

    def test_preserves_annotations(self) -> None:
        builder                = Subclasser()
        target         : str   = "TestCls"
        target_annots  : dict  = {"other": "float"}
        namespace      : dict  = {"__annotations__": target_annots}

        class Basic:
            val : int

        class ExSub(Basic):
            """ A Typically created subclass """
            other : float

        match ExSub.__annotations__:
            case dict() as annots if annots == target_annots:
                # Subclasses don't have their mro annotations
                ex_inst = ExSub()
                assert(ex_inst.__annotations__ == target_annots)
                assert(True)
            case other:
                assert(False), other

        dyn_subclass   = builder.make_subclass(target, Basic, namespace=namespace)
        match dyn_subclass:
            case type() as x:
                assert(x.__annotations__ == target_annots)
                inst = x()
                assert(inst.__annotations__ == target_annots)
            case x:
                assert(False), x

    def test_preserve_classvars(self) -> None:
        builder                = Subclasser()
        target         : str   = "TestCls"
        target_annots  : dict  = {"other":"ClassVar[float]"}
        namespace      : dict  = {
            "__annotations__": target_annots,
            "other" : 0.5,
        }

        class Basic:
            val : ClassVar[int] = 5

        class ExSub(Basic):
            """ A Typically created subclass """
            other : ClassVar[float] = 0.5

        match ExSub.__annotations__:
            case dict() as annots if annots == target_annots:
                # Subclasses don't have their mro annotations
                ex_inst = ExSub()
                assert(ex_inst.__annotations__ == target_annots)
                assert(ExSub.val == 5)
                assert(ExSub.other == 0.5)
                assert(ex_inst.val == 5)
                assert(ex_inst.other == 0.5)
            case other:
                assert(False), other

        dyn_subclass = builder.make_subclass(target, Basic, namespace=namespace)
        match dyn_subclass:
            case type() as x:
                assert(issubclass(x, Basic))
                assert(x.__annotations__ == target_annots)
                inst = x()
                assert(inst.__annotations__ == target_annots)
                assert(x.val == 5)
                assert(x.other == 0.5) # type: ignore[attr-defined]
                assert(inst.val == 5)
                assert(inst.other == 0.5) # type: ignore[attr-defined]
            case x:
                assert(False), x

    def test_classvars_and_slots_and_annotations(self) -> None:
        """ Check a built subclass can have:

        - classvars+values,
        - annotations
        - still no dict
        """

        builder                = Subclasser()
        target         : str   = "TestCls"
        target_annots  : dict  = {"other":"ClassVar[float]"}
        namespace      : dict  = {
            "__annotations__": target_annots,
            "other" : 0.5,
        }

        class Basic:
            __slots__ = ()
            val : ClassVar[int] = 5

        assert(Basic.val == 5)
        assert(not hasattr(Basic(), "__dict__"))

        class ExSub(Basic):
            """ A Typically created subclass """
            __slots__ = ()
            other : ClassVar[float] = 0.5

        assert(ExSub.val == 5)
        assert(ExSub.other == 0.5)
        match ExSub.__annotations__:
            case dict() as annots if annots == target_annots:
                # Subclasses don't have their mro annotations
                ex_inst = ExSub()
                assert(ex_inst.__annotations__ == target_annots)
                assert(ex_inst.val == 5)
                assert(ex_inst.other == 0.5)
                assert(not hasattr(ex_inst, "__dict__")), ex_inst.__dict__
            case other:
                assert(False), other

        dyn_subclass = builder.make_subclass(target, Basic, namespace=namespace)
        match dyn_subclass:
            case type() as x:
                assert(x.__annotations__ == target_annots)
                inst = x()
                assert(inst.__annotations__ == target_annots)
                assert(x.val == 5)
                assert(x.other == 0.5) # type: ignore[attr-defined]
                assert(inst.val == 5)
                assert(inst.other == 0.5) # type: ignore[attr-defined]
                assert(not hasattr(inst, "__dict__")), inst.__dict__
            case x:
                assert(False), x

    def test_pep0560(self):
        """ Check a subclass matches the example in pep0560
        """
        builder = Subclasser()

        class Basic:
            pass

        class Sub(builder.make_generic(Basic, int)):
            pass

        match Sub.__orig_bases__:
            case [val]:
                assert(val.__origin__ is Basic)
                assert(val.__args__ == (int,))
            case x:
                assert(False), x

class TestSubclasser_Annotation:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_annotation(self):
        builder        = Subclasser()

        class Basic:
            __slots__ = ()
            pass

        match builder.annotate(Basic, "test"):
            case type() as x:
                assert(issubclass(x, Basic))
                assert(Basic in x.mro())
                assert(x.__annotations__[AnnotationTarget] == "ClassVar[str]")
                inst = x()
                assert(inst.__annotations__[AnnotationTarget] == "ClassVar[str]")
                assert(not hasattr(inst, "__dict__"))
            case x:
                assert(False), x

